from __future__ import division, print_function
import abc
import numpy as np

from menpo.transform import Scale, AlignmentSimilarity
from menpo.model.modelinstance import PDM, OrthoPDM
from menpo.transform.modeldriven import ModelDrivenTransform, OrthoMDTransform
from menpo.visualize import print_dynamic, progress_bar_str

from menpo.fit.regression.trainer import (
    NonParametricRegressorTrainer, ParametricRegressorTrainer,
    SemiParametricClassifierBasedRegressorTrainer)
from menpo.fit.regression.regressionfunctions import mlr
from menpo.fit.regression.parametricfeatures import weights
from menpo.fitmultilevel.functions import mean_pointcloud
from menpo.fitmultilevel.featurefunctions import compute_features, sparse_hog

from .base import (SDMFitter, SDAAMFitter, SDCLMFitter)


class SDTrainer(object):
    r"""
    Mixin for Supervised Descent Trainers.

    Parameters
    ----------
    regression_type : `function`, or list of those, optional
        If list of length ``n_levels``, then a regression type is defined per
        level.

        If not a list or a list with length ``1``, then the specified rergession
        type will be applied to all pyramid levels.

        Examples of such closures can be found in :ref:`regression_functions`.

    regression_features :`` None`` or `string` or `function` or list of those, optional
        The features that are used during the regression.

        If `list`, a regression feature is defined per level.

        If not list or list with length ``1``, the specified regression feature
        will be used for all levels.

        Depending on the :map:`SDTrainer` object, this parameter can take
        different types.

    feature_type : ``None`` or `string` or `function` or list of those, optional
        Defines the features that will be extracted from the image.
        If list of length ``n_levels``, then a feature is defined per level.
        However, this requires that the ``pyramid_on_features`` flag is
        ``False``, so that the features are extracted at each level.
        The first element of the list specifies the features to be extracted
        at the lowest pyramidal level and so on.

        If not a list or a list with length ``1``:
            If ``pyramid_on_features`` is ``True``, the specified feature will
            be applied to the highest level.

            If ``pyramid_on_features`` is ``False``, the specified feature will
            be applied to all pyramid levels.

        Per level:
            If ``None``, the appearance model will be built using the original
            image representation, i.e. no features will be extracted from the
            original images.

            If `string`, image features will be computed by executing::

               feature_image = getattr(image.features, feature_type[level])()

            for each pyramidal level. For this to work properly each string
            needs to be one of menpo's standard image feature methods
            ('igo', 'hog', ...).
            Note that, in this case, the feature computation will be
            carried out using the default options.

        Non-default feature options and new experimental features can be
        defined using functions/closures. In this case, the functions must
        receive an image as input and return a particular feature
        representation of that image. For example::

            def igo_double_from_std_normalized_intensities(image)
                image = deepcopy(image)
                image.normalize_std_inplace()
                return image.feature_type.igo(double_angles=True)

        See :map:`ImageFeatures` for details more details on
        Menpo's standard image features and feature options.

    n_levels : `int` > ``0``, optional
        The number of multi-resolution pyramidal levels to be used.

    downscale : `float` >= ``1``, optional
        The downscale factor that will be used to create the different
        pyramidal levels. The scale factor will be::

            (downscale ** k) for k in range(n_levels)

    pyramid_on_features : `boolean`, optional
        If ``True``, the feature space is computed once at the highest scale and
        the Gaussian pyramid is applied on the feature images.

        If ``False``, the Gaussian pyramid is applied on the original images
        (intensities) and then features will be extracted at each level.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.
`
    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int` > ``0``, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    interpolator : `string`, optional
        The interpolator in use.

    Returns
    -------
    fitter : :map:`MultilevelFitter`
        The fitter object.

    Raises
    ------
    ValueError
        ``regression_type`` must be a `function` or a list of those
        containing ``1`` or ``n_levels`` elements
    ValueError
        n_levels must be `int` > ``0``
    ValueError
        ``downscale`` must be >= ``1``
    ValueError
        ``n_perturbations`` must be > 0
    ValueError
        ``feature_type`` must be a `string` or a `function` or a list of those
        containing ``1`` or ``n_levels`` elements
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, regression_type=mlr, regression_features=None,
                 feature_type=None, n_levels=3, downscale=1.2,
                 pyramid_on_features=True, noise_std=0.04, rotation=False,
                 n_perturbations=10, interpolator='scipy'):

        # check parameters
        regression_type_list = self.check_regression_type(regression_type,
                                                          n_levels)
        feature_type = self.check_feature_type(feature_type, n_levels,
                                               pyramid_on_features)
        self.check_n_levels(n_levels)
        self.check_downscale(downscale)
        self.check_n_permutations(n_perturbations)

        # store parameters
        self.regression_type = regression_type_list
        self.regression_features = regression_features
        self.feature_type = feature_type
        self.n_levels = n_levels
        self.downscale = downscale
        self.pyramid_on_features = pyramid_on_features
        self.noise_std = noise_std
        self.rotation = rotation
        self.n_perturbations = n_perturbations
        self.interpolator = interpolator

    def train(self, images, group=None, label='all', verbose=False, **kwargs):
        r"""
        Trains a Supervised Descent Regressor given a list of landmarked
        images.

        Parameters
        ----------
        images: list of :map:`MaskedImage`
            The set of landmarked images from which to build the SD.

        group : `string`, optional
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label: `string`, optional
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        verbose: `boolean`, optional
            Flag that controls information and progress printing.
        """
        if verbose:
            print_dynamic('- Computing reference shape')
        self.reference_shape = self._compute_reference_shape(images, group,
                                                             label)
        # store number of training images
        self.n_training_images = len(images)

        # normalize the scaling of all images wrt the reference_shape size
        self._rescale_reference_shape()
        normalized_images = self._normalization_wrt_reference_shape(
            images, group, label, self.reference_shape, self.interpolator,
            verbose=verbose)

        # create pyramid
        generators = self._create_pyramid(normalized_images, self.n_levels,
                                          self.downscale,
                                          self.pyramid_on_features,
                                          self.feature_type, verbose=verbose)

        # get feature images of all levels
        images = self._apply_pyramid_on_images(
            generators, self.n_levels, self.pyramid_on_features,
            self.feature_type, verbose=verbose)

        # this .reverse sets the lowest resolution as the first level
        images.reverse()

        # extract the ground truth shapes
        gt_shapes = [[i.landmarks[group][label].lms for i in img]
                     for img in images]

        # build the regressors
        if verbose:
            if self.n_levels > 1:
                print_dynamic('- Building regressors for each of the {} '
                              'pyramid levels\n'.format(self.n_levels))
            else:
                print_dynamic('- Building regressors\n')

        regressors = []
        # for each pyramid level (low --> high)
        for j, (level_images, level_gt_shapes) in enumerate(zip(images,
                                                                gt_shapes)):
            if verbose:
                if self.n_levels == 1:
                    print_dynamic('\n')
                elif self.n_levels > 1:
                    print_dynamic('\nLevel {}:\n'.format(j + 1))

            # build regressor
            trainer = self._set_regressor_trainer(j)
            if j == 0:
                regressor = trainer.train(level_images, level_gt_shapes,
                                          verbose=verbose, **kwargs)
            else:
                regressor = trainer.train(level_images, level_gt_shapes,
                                          level_shapes, verbose=verbose,
                                          **kwargs)

            if verbose:
                print_dynamic('- Perturbing shapes...')
            level_shapes = trainer.perturb_shapes(gt_shapes[0])

            regressors.append(regressor)
            count = 0
            total = len(regressors) * len(images[0]) * len(level_shapes[0])
            for k, r in enumerate(regressors):

                test_images = images[k]
                test_gt_shapes = gt_shapes[k]

                fitting_results = []
                for (i, gt_s, level_s) in zip(test_images, test_gt_shapes,
                                              level_shapes):
                    fr_list = []
                    for ls in level_s:
                        parameters = r.get_parameters(ls)
                        fr = r.fit(i, parameters)
                        fr.gt_shape = gt_s
                        fr_list.append(fr)
                        count += 1

                    fitting_results.append(fr_list)
                    if verbose:
                        print_dynamic('- Fitting shapes: {}'.format(
                            progress_bar_str((count + 1.) / total,
                                             show_bar=False)))

                level_shapes = [[Scale(self.downscale,
                                       n_dims=self.reference_shape.n_dims
                                       ).apply(fr.final_shape)
                                 for fr in fr_list]
                                for fr_list in fitting_results]

            if verbose:
                print_dynamic('- Fitting shapes: computing mean error...')
            mean_error = np.mean(np.array([fr.final_error
                                           for fr_list in fitting_results
                                           for fr in fr_list]))
            if verbose:
                print_dynamic("- Fitting shapes: mean error "
                              "is {0:.6f}.\n".format(mean_error))

        return self._build_supervised_descent_fitter(regressors)

    @classmethod
    def _normalization_wrt_reference_shape(cls, images, group, label,
                                           reference_shape, interpolator,
                                           verbose=False):
        r"""
        Function that normalizes the images sizes with respect to the reference
        shape (mean shape) scaling. This step is essential before building a
        deformable model.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images from which to build the model.

        group : `string`
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        reference_shape : :map:`PointCloud`
            The reference shape that is used to resize all training images to
            a consistent object size.

        interpolator : `string`
            The interpolator that should be used to perform the warps.

        verbose: bool, optional
            Flag that controls information and progress printing.

        Returns
        -------
        normalized_images : :map:`MaskedImage` list
            A list with the normalized images.
        """
        normalized_images = []
        for c, i in enumerate(images):
            if verbose:
                print_dynamic('- Normalizing images size: {}'.format(
                    progress_bar_str((c + 1.) / len(images),
                                     show_bar=False)))
            normalized_images.append(i.rescale_to_reference_shape(
                reference_shape, group=group, label=label,
                interpolator=interpolator))

        if verbose:
            print_dynamic('- Normalizing images size: Done\n')
        return normalized_images

    @classmethod
    def _create_pyramid(cls, images, n_levels, downscale, pyramid_on_features,
                        feature_type, verbose=False):
        r"""
        Function that creates a generator function for Gaussian pyramid. The
        pyramid can be created either on the feature space or the original
        (intensities) space.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        n_levels : `int`
            The number of multi-resolution pyramidal levels to be used.

        downscale : `float`
            The downscale factor that will be used to create the different
            pyramidal levels.

        pyramid_on_features : `boolean`
            If ``True``, the features are extracted at the highest level and the
            pyramid is created on the feature images.

            If ``False``, the pyramid is created on the original (intensities)
            space.

        feature_type : list with `string` or `function` or ``None``
            In case ``pyramid_on_features`` is ``True``, ``feature_type[0]``
            will be used as features type.

        verbose : `boolean`, Optional
            Flag that controls information and progress printing.

        Returns
        -------
        generator : `function`
            The generator function of the Gaussian pyramid.
        """
        if pyramid_on_features:
            # compute features at highest level
            feature_images = []
            for c, i in enumerate(images):
                if verbose:
                    print_dynamic('- Computing feature space: {}'.format(
                        progress_bar_str((c + 1.) / len(images),
                                         show_bar=False)))
                feature_images.append(compute_features(i, feature_type[0]))
            if verbose:
                print_dynamic('- Computing feature space: Done\n')

            # create pyramid on feature_images
            generator = [i.gaussian_pyramid(n_levels=n_levels,
                                            downscale=downscale)
                         for i in feature_images]
        else:
            # create pyramid on intensities images
            # features will be computed per level
            generator = [i.gaussian_pyramid(n_levels=n_levels,
                                            downscale=downscale)
                         for i in images]
        return generator

    @classmethod
    def _apply_pyramid_on_images(cls, generators, n_levels,
                                 pyramid_on_features, feature_type,
                                 verbose=False):
        r"""
        Function that applies the generators of a pyramid on images.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        generators : list of generator `function`
            The generator functions of the Gaussian pyramid for all images.

        n_levels: `int`
            The number of multi-resolution pyramidal levels to be used.

        pyramid_on_features: boolean
            If ``True``, the features are extracted at the highest level and the
            pyramid is created on the feature images.
            If ``False``, the pyramid is created on the original (intensities)
            space.

        feature_type: list of length ``n_levels`` with `string` or `function` or ``None``
            The feature type per level to be used in case
            ``pyramid_on_features`` is enabled.

        verbose: `boolean`, optional
            Flag that controls information and progress printing.

        Returns
        -------
        feature_images: list of lists of :map:`MaskedImage`
            The set of pyramidal images.
        """
        feature_images = []
        for j in range(n_levels):
            # since generators are built from highest to lowest level, the
            # parameters in form of list need to use a reversed index
            rj = n_levels - j - 1

            if verbose:
                level_str = '- Apply pyramid: '
                if n_levels > 1:
                    level_str = '- Apply pyramid: [Level {} - '.format(j + 1)

            current_images = []
            if pyramid_on_features:
                # features are already computed, so just call generator
                for c, g in enumerate(generators):
                    if verbose:
                        print_dynamic('{}Rescaling feature space - {}]'.format(
                            level_str,
                            progress_bar_str((c + 1.) / len(generators),
                                             show_bar=False)))
                    current_images.append(g.next())
            else:
                # extract features of images returned from generator
                for c, g in enumerate(generators):
                    if verbose:
                        print_dynamic('{}Computing feature space - {}]'.format(
                            level_str,
                            progress_bar_str((c + 1.) / len(generators),
                                             show_bar=False)))
                    current_images.append(compute_features(g.next(),
                                                           feature_type[rj]))
            feature_images.append(current_images)
        if verbose:
            print_dynamic('- Apply pyramid: Done\n')
        return feature_images

    @classmethod
    def check_feature_type(cls, feature_type, n_levels, pyramid_on_features):
        r"""
        Checks the feature type per level.

        If ``pyramid_on_features`` is ``False``, it must be a string or a
        function/closure or a list of those containing ``1`` or ``n_levels``
        elements.

        If ``pyramid_on_features`` is ``True``, it must be a string or a
        function/closure or a list of 1 of those.

        Parameters
        ----------
        feature_type : `string` or `function` or ``None`` or list of those
            The feature type to check.

        n_levels : `int`
            The number of pyramid levels.

        pyramid_on_features : `boolean`
            If ``True``, the pyramid will be applied to the feature image, so
            the user needs to define a single ``feature_type``.

            If ``False``, the pyramid will be applied to the intensities image
            and features will be extracted at each level, so the user can define
            a ``feature_type`` per level.

        Returns
        -------
        feature_type_list : l`ist`
            A list of feature types.

            If pyramid_on_features is ``True``, the list will have length ``1``.

            If pyramid_on_features is ``False``, the list will have length
            ``n_levels``.
        """
        if not pyramid_on_features:
            feature_type_str_error = ("feature_type must be a str or a "
                                      "function/closure or a list of "
                                      "those containing 1 or {} "
                                      "elements").format(n_levels)
            if not isinstance(feature_type, list):
                feature_type_list = [feature_type] * n_levels
            elif len(feature_type) == 1:
                feature_type_list = [feature_type[0]] * n_levels
            elif len(feature_type) == n_levels:
                feature_type_list = feature_type
            else:
                raise ValueError(feature_type_str_error)
        else:
            feature_type_str_error = ("pyramid_on_features is enabled so "
                                      "feature_type must be a str or a "
                                      "function/closure or a list "
                                      "containing 1 of those")
            if not isinstance(feature_type, list):
                feature_type_list = [feature_type]
            elif len(feature_type) == 1:
                feature_type_list = feature_type
            else:
                raise ValueError(feature_type_str_error)
        for ft in feature_type_list:
            if (ft is not None and not isinstance(ft, str)
                    and not hasattr(ft, '__call__')):
                        raise ValueError(feature_type_str_error)
        return feature_type_list

    @classmethod
    def check_regression_type(cls, regression_type, n_levels):
        r"""
        Checks the regression type (method) per level.

        It must be a function/closure or a list of those from the family of
        functions defined in :ref:`regression_functions`

        Parameters
        ----------
        regression_type : `function` or list of those
            The regression type to check.

        n_levels : `int`
            The number of pyramid levels.

        Returns
        -------
        regression_type_list : `list`
            A list of regression types that has length ``n_levels``.
        """
        regression_type_str_error = ("regression_type must be a "
                                     "function/closure or a list of "
                                     "those containing 1 or {} "
                                     "elements").format(n_levels)
        if not isinstance(regression_type, list):
            regression_type_list = [regression_type] * n_levels
        elif len(regression_type) == 1:
            regression_type_list = [regression_type[0]] * n_levels
        elif len(regression_type) == n_levels:
            regression_type_list = regression_type
        else:
            raise ValueError(regression_type_str_error)
        for rt in regression_type_list:
            if not hasattr(rt, '__call__'):
                raise ValueError(regression_type_str_error)
        return regression_type_list

    @classmethod
    def check_n_levels(cls, n_levels):
        r"""
        Checks the number of pyramid levels that must be `int` > ``0``.
        """
        if not isinstance(n_levels, int) or n_levels < 1:
            raise ValueError("n_levels must be int > 0")

    @classmethod
    def check_downscale(cls, downscale):
        r"""
        Checks the downscale factor of the pyramid that must be >= ``1``.
        """
        if downscale < 1:
            raise ValueError("downscale must be >= 1")

    @classmethod
    def check_n_permutations(cls, n_permutations):
        r"""
        Checks the n_permutations that must be > ``0``.
        """
        if n_permutations < 1:
            raise ValueError("n_permutations must be > 0")

    def check_regression_features(self, regression_features, n_levels):
        r"""
        Checks the regression features per level.

        Parameters
        ----------
        regression_features : `string` or `function` or ``None`` or list of those
            The regression features to check.

        n_levels : `int`
            The number of pyramid levels.

        Returns
        -------
        regression_features_list : `list`
            A list of regression features with length ``n_levels``.
        """
        pass

    @abc.abstractmethod
    def _compute_reference_shape(self, images, group, label):
        r"""
        Function that computes the reference shape, given a set of images.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        group : `string`
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        Returns
        -------
        reference_shape : :map:`PointCloud`
            The reference shape computed based on the given images shapes.
        """
        pass

    def _rescale_reference_shape(self):
        r"""
        Function that rescales the reference shape w.r.t. to
        ``normalization_diagonal`` parameter.
        """
        pass

    @abc.abstractmethod
    def _set_regressor_trainer(self, **kwargs):
        r"""
        Function that sets the regression object to be one from
        :map:`RegressorTrainer`,
        """
        pass

    @abc.abstractmethod
    def _build_supervised_descent_fitter(self, regressors):
        r"""
        Builds an SDM fitter object.

        Parameters
        ----------
        regressors : list of :map:`RegressorTrainer`
            The list of regressors.

        Returns
        -------
        fitter : :map:`SDMFitter`
            The SDM fitter object.
        """
        pass


class SDMTrainer(SDTrainer):
    r"""
    Class that trains Supervised Descent Method using Non-Parametric
    Regression.

    Parameters
    ----------
    regression_type : `function` or list of those, optional
        If list of length ``n_levels``, then a regression type is defined per
        level.

        If not a list or a list with length ``1``, then the specified regression
        type will be applied to all pyramid levels.

        The function/closures should be one of the methods defined in
        :ref:`regression_functions`

    regression_features: ``None`` or `string` or `function` or list of those, optional
        If list of length ``n_levels``, then a feature is defined per level.

        If not a list or a list with length 1, then the specified feature will
        be applied to all pyramid levels.

        Per level:
            If ``None``, no features are extracted, thus specified
            ``feature_type`` is used in the regressor.

            If `string`, image features will be computed using one of Menpo's
            standard image feature methods ('igo', 'hog', ...).

            Note that, in this case, the feature computation will be
            carried out using the default options.
            If function/closure, non-default feature options features can be
            defined using functions/closures. Such examples exist in
            :ref:`feature_functions`.

            It is recommended to set the desired features using this option,
            leaving ``feature_type`` equal to ``None``. This means that the
            images will remain in the intensities space and the features will
            be extracted by the regressor.

    patch_shape: tuple of `int`
        The shape of the patches used by the SDM.

    feature_type : `None` or `string` or `function` or list of those, optional
        If list of length ``n_levels``, then a feature is defined per level.
        However, this requires that the ``pyramid_on_features`` flag is
        ``False``, so that the features are extracted at each level.
        The first element of the list specifies the features to be extracted at
        the lowest pyramidal level and so on.

        If not a list or a list with length ``1``, then:
            If ``pyramid_on_features`` is ``True``, the specified feature will
            be applied to the highest level.

            If ``pyramid_on_features`` is ``False``, the specified feature will
            be applied to all pyramid levels.

        Per level:
            If ``None``, the appearance model will be built using the original
            image representation, i.e. no features will be extracted from the
            original images.

            If `string`, image features will be computed by executing::

               feature_image = getattr(image.features, feature_type[level])()

        for each pyramidal level. For this to work properly each string
        needs to be one of menpo's standard image feature methods
        ('igo', 'hog', ...).
        Note that, in this case, the feature computation will be
        carried out using the default options.

        Non-default feature options and new experimental features can be
        defined using functions/closures. In this case, the functions must
        receive an image as input and return a particular feature
        representation of that image. For example::

            def igo_double_from_std_normalized_intensities(image)
                image = deepcopy(image)
                image.normalize_std_inplace()
                return image.feature_type.igo(double_angles=True)

        See :map:`ImageFeatures` for details more details on
        Menpo's standard image features and feature options.

    n_levels : `int` > ``0``, optional
        The number of multi-resolution pyramidal levels to be used.

    downscale : `float` >= ``1``, optional
        The downscale factor that will be used to create the different
        pyramidal levels. The scale factor will be::

            (downscale ** k) for k in range(n_levels)

    pyramid_on_features : `boolean`, optional
        If ``True``, the feature space is computed once at the highest scale and
        the Gaussian pyramid is applied on the feature images.

        If ``False``, the Gaussian pyramid is applied on the original images
        (intensities) and then features will be extracted at each level.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        initial shape.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the initial shape.

    n_perturbations : `int` > ``0``, optional
        Defines the number of perturbations that will be applied to the shapes.

    normalization_diagonal : `int` >= ``20``, optional
        During training, all images are rescaled to ensure that the scale of
        their landmarks matches the scale of the mean shape.

        If `int`, it ensures that the mean shape is scaled so that the diagonal
        of the bounding box containing it matches the normalization_diagonal
        value.

        If ``None``, the mean shape is not rescaled.

        Note that, because the reference frame is computed from the mean
        landmarks, this kwarg also specifies the diagonal length of the
        reference frame (provided that features computation does not change
        the image size).

    interpolator : `string`, optional
        The interpolator in use.

    Raises
    ------
    ValueError
        ``regression_features`` must be ``None`` or a `string` or a `function`
        or a list of those containing 1 or ``n_level`` elements
    """
    def __init__(self, regression_type=mlr, regression_features=sparse_hog,
                 patch_shape=(16, 16), feature_type=None, n_levels=3,
                 downscale=1.5, pyramid_on_features=False, noise_std=0.04,
                 rotation=False, n_perturbations=10,
                 normalization_diagonal=None, interpolator='scipy'):
        # check regression features
        regression_features_list = self.check_regression_features(
            regression_features, n_levels)
        super(SDMTrainer, self).__init__(
            regression_type=regression_type,
            regression_features=regression_features_list,
            feature_type=feature_type, n_levels=n_levels, downscale=downscale,
            pyramid_on_features=pyramid_on_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations,
            interpolator=interpolator)
        self.patch_shape = patch_shape
        self.normalization_diagonal = normalization_diagonal
        self.pyramid_on_features = pyramid_on_features

    def _compute_reference_shape(self, images, group, label):
        r"""
        Function that computes the reference shape, given a set of images.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        group : `string`
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        Returns
        -------
        reference_shape : :map:`PointCloud`
            The reference shape computed based on the given images.
        """
        shapes = [i.landmarks[group][label].lms for i in images]
        return mean_pointcloud(shapes)

    def _rescale_reference_shape(self):
        r"""
        Function that rescales the reference shape w.r.t. to
        ``normalization_diagonal`` parameter.
        """
        if self.normalization_diagonal:
            x, y = self.reference_shape.range()
            scale = self.normalization_diagonal / np.sqrt(x**2 + y**2)
            Scale(scale, self.reference_shape.n_dims).apply_inplace(
                self.reference_shape)

    def _set_regressor_trainer(self, level):
        r"""
        Function that sets the regression class to be the
        :map:`NonParametricRegressorTrainer`.

        Parameters
        ----------
        level : `int`
            The scale level.

        Returns
        -------
        trainer : :map:`NonParametricRegressorTrainer`
            The regressor object.
        """
        return NonParametricRegressorTrainer(
            self.reference_shape, regression_type=self.regression_type[level],
            regression_features=self.regression_features[level],
            patch_shape=self.patch_shape, noise_std=self.noise_std,
            rotation=self.rotation, n_perturbations=self.n_perturbations)

    def _build_supervised_descent_fitter(self, regressors):
        r"""
        Builds an SDM fitter object.

        Parameters
        ----------
        regressors : list of :map:`RegressorTrainer`
            The list of regressors.

        Returns
        -------
        fitter : :map:`SDMFitter`
            The SDM fitter object.
        """
        return SDMFitter(regressors, self.n_training_images, self.feature_type,
                         self.reference_shape, self.downscale,
                         self.pyramid_on_features, self.interpolator)

    def check_regression_features(self, regression_features, n_levels):
        r"""
        Checks the regression features per level.

        It must be a `string` or a function` or a list of those
        containing ``1`` or ``n_levels`` elements.

        Parameters
        ----------
        regression_features : `string` or `function` or ``None`` or list of those
            The regression features to check.

        n_levels : `int`
            The number of pyramid levels.

        Returns
        -------
        regression_features_list : `list`
            A list of regression features with length ``n_levels``.
        """
        regression_features_str_error = ("regression_features must be None "
                                         "or a str or a function/closure or "
                                         "a list of those containing 1 or {} "
                                         "elements").format(n_levels)
        if not isinstance(regression_features, list):
            regression_features_list = [regression_features] * n_levels
        elif len(regression_features) == 1:
            regression_features_list = [regression_features[0]] * n_levels
        elif len(regression_features) == n_levels:
            regression_features_list = regression_features
        else:
            raise ValueError(regression_features_str_error)
        for rf in regression_features_list:
            if (rf is not None and not isinstance(rf, str)
                    and not hasattr(rf, '__call__')):
                        raise ValueError(regression_features_str_error)
        return regression_features_list


class SDAAMTrainer(SDTrainer):
    r"""
    Class that trains Supervised Descent Regressor for a given Active
    Appearance Model, thus uses Parametric Regression.

    Parameters
    ----------
    aam : :map:`AAM`
        The trained AAM object.

    regression_type: `function` or list of those, optional
        If list of length ``n_levels``, then a regression type is defined per
        level.

        If not a list or a list with length ``1``, then the specified regression
        type will be applied to all pyramid levels.

        The function/closures should be one of the methods defined in
        :ref:`regression_functions`

    regression_features: `function` or list of those, optional
        If list of length ``n_levels``, then a feature is defined per level.

        If not a list or a list with length ``1``, then the specified feature
        will be applied to all pyramid levels.

        The function/closures should be one of the methods defined in
        :ref:`parametric_features`.

    noise_std : `float`, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int` > ``0``, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    update : {'additive', 'compositional'}
        Defines the way that the warp will be updated.

    md_transform: :map:`ModelDrivenTransform`, optional
        The model driven transform class to be used.

    global_transform : :map:`Affine`, optional
        The global transform class to be used by the previous
        ``md_transform_cls``. Currently, only
        :map:`AlignmentSimilarity` is supported.

    n_shape : `int` > ``1`` or ``0`` <= `float` <= ``1`` or ``None``, or a list of those, optional
        The number of shape components to be used per fitting level.

        If list of length ``n_levels``, then a number of components is defined
        per level. The first element of the list corresponds to the lowest
        pyramidal level and so on.

        If not a list or a list with length ``1``, then the specified number of
        components will be used for all levels.

        Per level:
            If ``None``, all the available shape components
            (``n_active_components``)will be used.

            If `int` > ``1``, a specific number of shape components is
            specified.

            If ``0`` <= `float` <= ``1``, it specifies the variance percentage
            that is captured by the components.

    n_appearance : `int` > ``1`` or ``0`` <= `float` <= ``1`` or ``None``, or a list of those, optional
        The number of appearance components to be used per fitting level.

        If list of length ``n_levels``, then a number of components is defined
        per level. The first element of the list corresponds to the lowest
        pyramidal level and so on.

        If not a list or a list with length 1, then the specified number of
        components will be used for all levels.

        Per level:
            If ``None``, all the available appearance components
            (``n_active_components``) will be used.
            
            If `int > ``1``, a specific number of appearance components is
            specified.
            
            If ``0`` <= `float` <= ``1``, it specifies the variance percentage
            that is captured by the components.

    Raises
    -------
    ValueError
        n_shape can be an integer or a float or None or a list containing 1
        or ``n_levels`` of those
    ValueError
        n_appearance can be an integer or a float or None or a list containing
        1 or ``n_levels`` of those
    ValueError
        ``regression_features`` must be a `function` or a list of those
        containing ``1`` or ``n_levels`` elements
    """
    def __init__(self, aam, regression_type=mlr, regression_features=weights,
                 noise_std=0.04, rotation=False, n_perturbations=10,
                 update='compositional', md_transform=OrthoMDTransform,
                 global_transform=AlignmentSimilarity, n_shape=None,
                 n_appearance=None):
        # check regression features
        regression_features_list = self.check_regression_features(
            regression_features, aam.n_levels)
        super(SDAAMTrainer, self).__init__(
            regression_type=regression_type,
            regression_features=regression_features_list,
            feature_type=aam.feature_type, n_levels=aam.n_levels,
            downscale=aam.downscale,
            pyramid_on_features=aam.pyramid_on_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations,
            interpolator=aam.interpolator)
        self.aam = aam
        self.update = update
        self.md_transform = md_transform
        self.global_transform = global_transform

        # check n_shape parameter
        if n_shape is not None:
            if type(n_shape) is int or type(n_shape) is float:
                for sm in self.aam.shape_models:
                    sm.n_active_components = n_shape
            elif len(n_shape) == 1 and self.aam.n_levels > 1:
                for sm in self.aam.shape_models:
                    sm.n_active_components = n_shape[0]
            elif len(n_shape) == self.aam.n_levels:
                for sm, n in zip(self.aam.shape_models, n_shape):
                    sm.n_active_components = n
            else:
                raise ValueError('n_shape can be an integer or a float, '
                                 'an integer or float list containing 1 '
                                 'or {} elements or else '
                                 'None'.format(self.aam.n_levels))

        # check n_appearance parameter
        if n_appearance is not None:
            if type(n_appearance) is int or type(n_appearance) is float:
                for am in self.aam.appearance_models:
                    am.n_active_components = n_appearance
            elif len(n_appearance) == 1 and self.aam.n_levels > 1:
                for am in self.aam.appearance_models:
                    am.n_active_components = n_appearance[0]
            elif len(n_appearance) == self.aam.n_levels:
                for am, n in zip(self.aam.appearance_models, n_appearance):
                    am.n_active_components = n
            else:
                raise ValueError('n_appearance can be an integer or a float, '
                                 'an integer or float list containing 1 '
                                 'or {} elements or else '
                                 'None'.format(self.aam.n_levels))

    def _compute_reference_shape(self, images, group, label):
        r"""
        Function that returns the reference shape computed during AAM building.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        group : `string`
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        Returns
        -------
        reference_shape : :map:`PointCloud`
            The reference shape computed based on.
        """
        return self.aam.reference_shape

    def _normalize_object_size(self, images, group, label):
        r"""
        Function that normalizes the images sizes with respect to the reference
        shape (mean shape) scaling.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images from which to build the model.

        group : `string`
            The key of the landmark set that should be used. If ```None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        Returns
        -------
        normalized_images : :map:`MaskedImage` list
            A list with the normalized images.
        """
        return [i.rescale_to_reference_shape(self.reference_shape,
                                             group=group, label=label,
                                             interpolator=self.interpolator)
                for i in images]

    def _set_regressor_trainer(self, level):
        r"""
        Function that sets the regression class to be the
        :map:`ParametricRegressorTrainer`.

        Parameters
        ----------
        level : `int`
            The scale level.

        Returns
        -------
        trainer: :map:`ParametricRegressorTrainer`
            The regressor object.
        """
        am = self.aam.appearance_models[level]
        sm = self.aam.shape_models[level]

        if self.md_transform is not ModelDrivenTransform:
            md_transform = self.md_transform(
                sm, self.aam.transform, self.global_transform,
                source=am.mean.landmarks['source'].lms)
        else:
            md_transform = self.md_transform(
                sm, self.aam.transform,
                source=am.mean.landmarks['source'].lms)

        return ParametricRegressorTrainer(
            am, md_transform, self.reference_shape,
            regression_type=self.regression_type[level],
            regression_features=self.regression_features[level],
            update=self.update, noise_std=self.noise_std,
            rotation=self.rotation, n_perturbations=self.n_perturbations,
            interpolator=self.interpolator)

    def _build_supervised_descent_fitter(self, regressors):
        r"""
        Builds an SDM fitter object for AAMs.

        Parameters
        ----------
        regressors : :map:`RegressorTrainer`
            The regressor to build with.

        Returns
        -------
        fitter : :map:`SDAAMFitter`
            The SDM fitter object.
        """
        return SDAAMFitter(self.aam, regressors, self.n_training_images)

    def check_regression_features(self, regression_features, n_levels):
        r"""
        Checks the regression features per level.

        It must be a function/closure or a list of those containing ``1`` or
        ``n_levels`` elements.

        Parameters
        ----------
        regression_features : `function` or list of those
            The regression features to check.

        n_levels : `int`
            The number of pyramid levels.

        Returns
        -------
        regression_features_list : `list`
            A list of regression features with length ``n_levels``.
        """
        regression_features_str_error = ("regression_features must be "
                                         "a function/closure or "
                                         "a list of those containing 1 or {} "
                                         "elements").format(n_levels)
        if not isinstance(regression_features, list):
            regression_features_list = [regression_features] * n_levels
        elif len(regression_features) == 1:
            regression_features_list = [regression_features[0]] * n_levels
        elif len(regression_features) == n_levels:
            regression_features_list = regression_features
        else:
            raise ValueError(regression_features_str_error)
        for rf in regression_features_list:
            if not hasattr(rf, '__call__'):
                raise ValueError(regression_features_str_error)
        return regression_features_list


class SDCLMTrainer(SDTrainer):
    r"""
    Class that trains Supervised Descent Regressor for a given Constrained
    Local Model, thus uses Semi Parametric Classifier-Based Regression.

    Parameters
    ----------
    clm : :map:`CLM`
        The trained CLM object.

    regression_type: `function` or list of those, optional
        If list of length ``n_levels``, then a regression type is defined per
        level.

        If not a list or a list with length ``1``, then the specified regression
        type will be applied to all pyramid levels.

        The function/closures should be one of the methods defined in
        :ref:`regression_functions`.

    noise_std: float, optional
        The standard deviation of the gaussian noise used to produce the
        training shapes.

    rotation : `boolean`, optional
        Specifies whether ground truth in-plane rotation is to be used
        to produce the training shapes.

    n_perturbations : `int` > ``0``, optional
        Defines the number of perturbations that will be applied to the
        training shapes.

    pdm_transform : :map:`ModelDrivenTransform`, optional
        The point distribution transform class to be used.

    global_transform : :map:`Affine`, optional
        The global transform class to be used by the previous
        ``md_transform_cls``. Currently, only
        :map:`AlignmentSimilarity` is supported.

    n_shape : `int` > ``1`` or ``0`` <= `float` <= ``1`` or ``None``, or a list of those, optional
        The number of shape components to be used per fitting level.

        If list of length ``n_levels``, then a number of components is defined
        per level. The first element of the list corresponds to the lowest
        pyramidal level and so on.

        If not a list or a list with length ``1``, then the specified number of
        components will be used for all levels.

        Per level:
            If ``None``, all the available shape components
            (``n_active_components``) will be used.

            If `int` > ``1``, a specific number of shape components is
            specified.
            
            If ``0`` <= `float` <= ``1``, it specifies the variance percentage
            that is captured by the components.

    Raises
    -------
    ValueError
        ``n_shape`` can be an integer or a `float` or ``None`` or a list
        containing ``1`` or ``n_levels`` of those.
    """
    def __init__(self, clm, regression_type=mlr, noise_std=0.04,
                 rotation=False, n_perturbations=10, pdm_transform=OrthoPDM,
                 global_transform=AlignmentSimilarity, n_shape=None):
        super(SDCLMTrainer, self).__init__(
            regression_type=regression_type,
            regression_features=[None] * clm.n_levels,
            feature_type=clm.feature_type, n_levels=clm.n_levels,
            downscale=clm.downscale,
            pyramid_on_features=clm.pyramid_on_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations,
            interpolator=clm.interpolator)
        self.clm = clm
        self.patch_shape = clm.patch_shape
        self.pdm_transform = pdm_transform
        self.global_transform = global_transform

        # check n_shape parameter
        if n_shape is not None:
            if type(n_shape) is int or type(n_shape) is float:
                for sm in self.clm.shape_models:
                    sm.n_active_components = n_shape
            elif len(n_shape) == 1 and self.clm.n_levels > 1:
                for sm in self.clm.shape_models:
                    sm.n_active_components = n_shape[0]
            elif len(n_shape) == self.clm.n_levels:
                for sm, n in zip(self.clm.shape_models, n_shape):
                    sm.n_active_components = n
            else:
                raise ValueError('n_shape can be an integer or a float or None'
                                 'or a list containing 1 or {} of '
                                 'those'.format(self.clm.n_levels))

    def _compute_reference_shape(self, images, group, label):
        r"""
        Function that returns the reference shape computed during CLM building.

        Parameters
        ----------
        images : list of :map:`MaskedImage`
            The set of landmarked images.

        group : `string`
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`
            The label of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        Returns
        -------
        reference_shape : :map:`PointCloud`
            The reference shape.
        """
        return self.clm.reference_shape

    def _set_regressor_trainer(self, level):
        r"""
        Function that sets the regression class to be the
        :map:`SemiParametricClassifierBasedRegressorTrainer`

        Parameters
        ----------
        level : `int`
            The scale level.

        Returns
        -------
        trainer: :map:`SemiParametricClassifierBasedRegressorTrainer`
            The regressor object.
        """
        clfs = self.clm.classifiers[level]
        sm = self.clm.shape_models[level]

        if self.pdm_transform is not PDM:
            pdm_transform = self.pdm_transform(sm, self.global_transform)
        else:
            pdm_transform = self.pdm_transform(sm)

        return SemiParametricClassifierBasedRegressorTrainer(
            clfs, pdm_transform, self.reference_shape,
            regression_type=self.regression_type[level],
            patch_shape=self.patch_shape, update='additive',
            noise_std=self.noise_std, rotation=self.rotation,
            n_perturbations=self.n_perturbations)

    def _build_supervised_descent_fitter(self, regressors):
        r"""
        Builds an SDM fitter object for CLMs.

        Parameters
        ----------
        regressors : :map:`RegressorTrainer`
            Regressor to train with.

        Returns
        -------
        fitter : :map:`SDCLMFitter`
            The SDM fitter object.
        """
        return SDCLMFitter(self.clm, regressors, self.n_training_images)



def compute_features(image, feature_type):
    r"""
    Computes a particular feature representation of the given images.

    Parameters
    ----------
    image : :map:`MaskedImage`
        The original image from which the features will be computed.

    feature_type : `string` or `function`
        If ``None``, no feature representation will be computed from the
        original image.

        If `string`, the feature representation will be extracted by
        executing::

            feature_image = getattr(image.features, feature_type)()

        For this to work properly feature_type needs to be one of
        Menpo's standard image feature methods. Note that, in this case,
        the feature computation will be carried out using its default
        options.

        Non-default feature options and new experimental feature can be
        used by defining a closure. In this case, the closure must define a
        function that receives as an input an image and returns a
        particular feature representation of that image. For example::

            def igo_double_from_std_normalized_intensities(image)
                image = deepcopy(image)
                image.normalize_std_inplace()
                return image.feature_type.igo(double_angles=True)

        See :map:`ImageFeatures` for details more details on
        Menpo's standard image features and feature options.

    Returns
    -------
    feature_image : :map:`MaskedImage`
        The resulting feature image.
    """
    if feature_type is not None:
        if isinstance(feature_type, str):
            image = getattr(image.features, feature_type)()
        elif hasattr(feature_type, '__call__'):
            image = feature_type(image)
        else:
            raise ValueError("feature_type can only be: (1) None, "
                             "(2) a string defining one of Menpo's standard "
                             "image feature_type ('hog', 'igo', etc) "
                             "or (3) a closure defining a non-standard "
                             "feature computation")
    return image


def sparse_hog(image):
    r"""
    Non-standard feature function that computes sparse HOGs.
    """
    return image.features.hog(mode='sparse', constrain_landmarks=True)


def double_igo(image):
    r"""
    Non-standard feature function that computes IGOs with double angles.
    """
    return image.features.igo(double_angles=True)



from __future__ import division
import abc
import numpy as np
from copy import deepcopy

from menpo.fit.base import Fitter
from menpo.transform import AlignmentAffine, Scale
from menpo.fitmultilevel.featurefunctions import compute_features

from .fittingresult import MultilevelFittingResult
from .functions import noisy_align, align_shape_with_bb


class MultilevelFitter(Fitter):
    r"""
    Mixin that all MultilevelFitter objects must implement.
    """

    @abc.abstractproperty
    def reference_shape(self):
        r"""
        Returns the reference shape. Typically, the mean of the shape model.
        """
        pass

    @abc.abstractproperty
    def feature_type(self):
        r"""
        Defines the feature computation function.
        """
        pass

    @abc.abstractproperty
    def n_levels(self):
        r"""
        Returns the number of levels used by the fitter.
        """
        pass

    @abc.abstractproperty
    def downscale(self):
        r"""
        Returns the downscale factor used by the fitter.
        """
        pass

    @abc.abstractproperty
    def pyramid_on_features(self):
        r"""
        Returns True if the pyramid is computed on the feature image and False
        if it is computed on the original (intensities) image and features are
        extracted at each level.
        """
        pass

    @abc.abstractproperty
    def interpolator(self):
        r"""
        Returns the type of interpolator used by the fitter.
        """
        pass

    def fit(self, image, initial_shape, max_iters=50, gt_shape=None,
            error_type='me_norm', verbose=False, view=False, **kwargs):
        r"""
        Fits a single image.

        Parameters
        -----------
        image: :class:`menpo.image.masked.MaskedImage`
            The image to be fitted.
        initial_shape: :class:`menpo.shape.PointCloud`
            The initial shape estimate from which the fitting procedure
            will start.
        max_iters: int or list, optional
            The maximum number of iterations.
            If int, then this will be the overall maximum number of iterations
            for all the pyramidal levels.
            If list, then a maximum number of iterations is specified for each
            pyramidal level.

            Default: 50
        gt_shape: PointCloud
            The groundtruth shape of the image.

            Default: None
        error_type: 'me_norm', 'me' or 'rmse', optional.
            Specifies the way in which the error between the fitted and
            ground truth shapes is to be computed.

            Default: 'me_norm'
        verbose: boolean, optional
            If True, it prints information related to the fitting results (such
            as: final error, convergence, ...).

            Default: False
        view: boolean, optional
            If True, the final fitting result will be visualized.

            Default: False

        **kwargs:

        Returns
        -------
        fitting_list: :map:`FittingResultList`
            A fitting result object.
        """
        # copy image
        image = deepcopy(image)

        # generate image pyramid
        images = self._prepare_image(image, initial_shape, gt_shape=gt_shape)

        # get ground truth shape per level
        if gt_shape:
            gt_shapes = [i.landmarks['gt_shape'].lms for i in images]
        else:
            gt_shapes = None

        # get initial shape per level
        initial_shapes = [i.landmarks['initial_shape'].lms for i in images]

        affine_correction = AlignmentAffine(initial_shapes[-1], initial_shape)

        # execute multilevel fitting
        fitting_results = self._fit(images, initial_shapes[0],
                                    max_iters=max_iters,
                                    gt_shapes=gt_shapes, **kwargs)

        # store result
        multilevel_fitting_result = self._create_fitting_result(
            image, fitting_results, affine_correction, gt_shape=gt_shape,
            error_type=error_type)

        if verbose:
            print multilevel_fitting_result
        if view:
            multilevel_fitting_result.view_final_fitting(new_figure=True)

        return multilevel_fitting_result

    def perturb_shape(self, gt_shape, noise_std=0.04, rotation=False):
        r"""
        Generates an initial shape by adding gaussian noise to the perfect
        similarity alignment between the ground truth and reference_shape.

        Parameters
        -----------
        gt_shape: :class:`menpo.shape.PointCloud`
            The ground truth shape.
        noise_std: float, optional
            The standard deviation of the gaussian noise used to produce the
            initial shape.

            Default: 0.04
        rotation: boolean, optional
            Specifies whether ground truth in-plane rotation is to be used
            to produce the initial shape.

            Default: False

        Returns
        -------
        initial_shape: :class:`menpo.shape.PointCloud`
            The initial shape.
        """
        reference_shape = self.reference_shape
        return noisy_align(reference_shape, gt_shape, noise_std=noise_std,
                           rotation=rotation).apply(reference_shape)

    def obtain_shape_from_bb(self, bounding_box):
        r"""
        Generates an initial shape given a bounding box detection.

        Parameters
        -----------
        bounding_box: (2, 2) ndarray
            The bounding box specified as:

                np.array([[x_min, y_min], [x_max, y_max]])

        Returns
        -------
        initial_shape: :class:`menpo.shape.PointCloud`
            The initial shape.
        """
        reference_shape = self.reference_shape
        return align_shape_with_bb(reference_shape,
                                   bounding_box).apply(reference_shape)

    def _prepare_image(self, image, initial_shape, gt_shape=None):
        r"""
        The image is first rescaled wrt the reference_landmarks and then the
        gaussian pyramid is computed. Depending on the pyramid_on_features
        flag, the pyramid is either applied on the feature image or
        features are extracted at each pyramidal level.

        Parameters
        ----------
        image: :class:`menpo.image.MaskedImage`
            The image to be fitted.
        initial_shape: class:`menpo.shape.PointCloud`
            The initial shape from which the fitting will start.
        gt_shape: class:`menpo.shape.PointCloud`, optional
            The original ground truth shape associated to the image.

            Default: None

        Returns
        -------
        images: list of :class:`menpo.image.masked.MaskedImage`
            List of images, each being the result of applying the pyramid.
        """
        # rescale image wrt the scale factor between reference_shape and
        # initial_shape
        image.landmarks['initial_shape'] = initial_shape
        image = image.rescale_to_reference_shape(
            self.reference_shape, group='initial_shape',
            interpolator=self.interpolator)

        # attach given ground truth shape
        if gt_shape:
            image.landmarks['gt_shape'] = gt_shape

        # apply pyramid
        if self.n_levels > 1:
            if self.pyramid_on_features:
                # compute features at highest level
                feature_image = compute_features(image, self.feature_type[0])

                # apply pyramid on feature image
                pyramid = feature_image.gaussian_pyramid(
                    n_levels=self.n_levels, downscale=self.downscale)

                # get rescaled feature images
                images = list(pyramid)
            else:
                # create pyramid on intensities image
                pyramid = image.gaussian_pyramid(
                    n_levels=self.n_levels, downscale=self.downscale)

                # compute features at each level
                images = [compute_features(
                    i, self.feature_type[self.n_levels - j - 1])
                    for j, i in enumerate(pyramid)]
            images.reverse()
        else:
            images = [compute_features(image, self.feature_type[0])]
        return images

    def _create_fitting_result(self, image, fitting_results, affine_correction,
                               gt_shape=None, error_type='me_norm'):
        r"""
        Creates the :class: `menpo.aam.fitting.MultipleFitting` object
        associated with a particular Fitter object.

        Parameters
        -----------
        image: :class:`menpo.image.masked.MaskedImage`
            The original image to be fitted.
        fitting_results: :class:`menpo.fit.fittingresult.FittingResultList`
            A list of basic fitting objects containing the state of the
            different fitting levels.
        affine_correction: :class: `menpo.transforms.affine.Affine`
            An affine transform that maps the result of the top resolution
            fitting level to the space scale of the original image.
        gt_shape: class:`menpo.shape.PointCloud`, optional
            The ground truth shape associated to the image.

            Default: None
        error_type: 'me_norm', 'me' or 'rmse', optional
            Specifies the way in which the error between the fitted and
            ground truth shapes is to be computed.

            Default: 'me_norm'

        Returns
        -------
        fitting: :class:`menpo.fitmultilevel.fittingresult.MultilevelFittingResult`
            The fitting object that will hold the state of the fitter.
        """
        return MultilevelFittingResult(image, self, fitting_results,
                                       affine_correction, gt_shape=gt_shape,
                                       error_type=error_type)

    def _fit(self, images, initial_shape, gt_shapes=None, max_iters=50,
             **kwargs):
        r"""
        Fits the fitter to the multilevel pyramidal images.

        Parameters
        -----------
        images: :class:`menpo.image.masked.MaskedImage` list
            The images to be fitted.
        initial_shape: :class:`menpo.shape.PointCloud`
            The initial shape from which the fitting will start.
        gt_shapes: :class:`menpo.shape.PointCloud` list, optional
            The original ground truth shapes associated to the multilevel
            images.

            Default: None
        max_iters: int or list, optional
            The maximum number of iterations.
            If int, then this will be the overall maximum number of iterations
            for all the pyramidal levels.
            If list, then a maximum number of iterations is specified for each
            pyramidal level.

            Default: 50

        Returns
        -------
        fitting_results: :class:`menpo.fit.fittingresult.FittingResult` list
            The fitting object containing the state of the whole fitting
            procedure.
        """
        shape = initial_shape
        n_levels = self.n_levels

        # check max_iters parameter
        if type(max_iters) is int:
            max_iters = [np.round(max_iters/n_levels)
                         for _ in range(n_levels)]
        elif len(max_iters) == 1 and n_levels > 1:
            max_iters = [np.round(max_iters[0]/n_levels)
                         for _ in range(n_levels)]
        elif len(max_iters) != n_levels:
            raise ValueError('max_iters can be integer, integer list '
                             'containing 1 or {} elements or '
                             'None'.format(self.n_levels))

        # fitting
        gt = None
        fitting_results = []
        for j, (i, f, it) in enumerate(zip(images, self._fitters, max_iters)):
            if gt_shapes is not None:
                gt = gt_shapes[j]

            parameters = f.get_parameters(shape)
            fitting_result = f.fit(i, parameters, gt_shape=gt, max_iters=it,
                                   **kwargs)
            fitting_results.append(fitting_result)

            shape = fitting_result.final_shape
            Scale(self.downscale, n_dims=shape.n_dims).apply_inplace(shape)

        return fitting_results



from __future__ import division
import numpy as np
from scipy.stats import multivariate_normal

from menpo.fit.base import Fitter
from menpo.fit.fittingresult import SemiParametricFittingResult
from menpo.fitmultilevel.functions import build_sampling_grid


# TODO: incorporate different residuals
# TODO: generalize transform prior, and map the changes to LK methods
class GradientDescent(Fitter):
    r"""
    Abstract Interface for defining Gradient Descent based fitting algorithms
    for Constrained Local Models [1]_.

    Parameters
    ----------
    classifiers : `list` of ``classifier_closure``
        The list containing the classifier that will produce the response
        maps for each landmark point.

    patch_shape : `tuple` of `int`
        The shape of the patches used to train the classifiers.

    transform : :map:`GlobalPDM` or subclass
        The global point distribution model to be used.

        .. note::

            Only :map:`GlobalPDM` and its subclasses are supported.
            :map:`PDM` is not supported at the moment.

    eps : `float`, optional
        The convergence value. When calculating the level of convergence, if
        the norm of the delta parameter updates is less than ``eps``, the
        algorithm is considered to have converged.

    References
    ----------
    .. [1] J. Saragih, S. Lucey and J. Cohn, ''Deformable Model Fitting by
    Regularized Landmark Mean-Shifts", International Journal of Computer
    Vision (IJCV), 2010.
    """
    def __init__(self, classifiers, patch_shape, pdm, eps=10**-10):
        self.classifiers = classifiers
        self.patch_shape = patch_shape
        self.transform = pdm
        self.eps = eps
        # pre-computations
        self._set_up()

    def _create_fitting_result(self, image, parameters, gt_shape=None):
        return SemiParametricFittingResult(
            image, self, parameters=[parameters], gt_shape=gt_shape)

    def fit(self, image, initial_parameters, gt_shape=None, **kwargs):
        self.transform.from_vector_inplace(initial_parameters)
        return Fitter.fit(self, image, initial_parameters, gt_shape=gt_shape,
                          **kwargs)

    def get_parameters(self, shape):
        self.transform.set_target(shape)
        return self.transform.as_vector()


class RegularizedLandmarkMeanShift(GradientDescent):
    r"""
    Implementation of the Regularized Landmark Mean-Shifts algorithm for
    fitting Constrained Local Models described in [1]_.

    Parameters
    ----------
    classifiers : `list` of ``classifier_closure``
        The list containing the classifier that will produce the response
        maps for each landmark point.

    patch_shape : `tuple` of `int`
        The shape of the patches used to train the classifiers.

    transform : :map:`GlobalPDM` or subclass
        The global point distribution model to be used.

        .. note::

            Only :map:`GlobalPDM` and its subclasses are supported.
            :map:`PDM` is not supported at the moment.

    eps : `float`, optional
        The convergence value. When calculating the level of convergence, if
        the norm of the delta parameter updates is less than ``eps``, the
        algorithm is considered to have converged.

    scale: `float`, optional
        Constant value that will be multiplied to the `noise_variance` of
        the pdm in order to compute the covariance of the KDE
        approximation.

     References
    ----------
    .. [1] J. Saragih, S. Lucey and J. Cohn, ''Deformable Model Fitting by
    Regularized Landmark Mean-Shifts", International Journal of Computer
    Vision (IJCV), 2010.
    """
    def __init__(self, classifiers, patch_shape, pdm, eps=10**-10, scale=10):
        self.scale = scale
        super(RegularizedLandmarkMeanShift, self).__init__(
            classifiers, patch_shape, pdm, eps=eps)

    @property
    def algorithm(self):
        return 'RLMS'

    def _set_up(self):
        # Build the sampling grid associated to the patch shape
        self._sampling_grid = build_sampling_grid(self.patch_shape)
        # Define the 2-dimensional gaussian distribution
        mean = np.zeros(self.transform.n_dims)
        covariance = self.scale * self.transform.model.noise_variance
        mvn = multivariate_normal(mean=mean, cov=covariance)
        # Compute Gaussian-KDE grid
        self._kernel_grid = mvn.pdf(self._sampling_grid)

        # Jacobian
        self._J = self.transform.d_dp([])

        # Prior
        sim_prior = np.zeros((4,))
        pdm_prior = 1 / self.transform.model.eigenvalues
        self._J_prior = np.hstack((sim_prior, pdm_prior))

        # Inverse Hessian
        H = np.einsum('ijk, ilk -> jl', self._J, self._J)
        self._inv_H = np.linalg.inv(np.diag(self._J_prior) + H)

    def _fit(self, fitting_result, max_iters=20):
        # Initial error > eps
        error = self.eps + 1
        image = fitting_result.image
        target = self.transform.target
        n_iters = 0

        max_x = image.shape[0] - 1
        max_y = image.shape[1] - 1

        image_pixels = np.reshape(image.pixels, (-1, image.n_channels))
        response_image = np.zeros((image.shape[0], image.shape[1],
                                   target.n_points))

        # Compute response maps
        for j, clf in enumerate(self.classifiers):
            response_image[:, :, j] = np.reshape(clf(image_pixels),
                                                 image.shape)

        while n_iters < max_iters and error > self.eps:

            mean_shift_target = np.zeros_like(target.points)

            # Compute mean-shift vectors
            for j, point in enumerate(target.points):

                patch_grid = (self._sampling_grid +
                              np.round(point[None, None, ...]).astype(int))

                x = patch_grid[:, :, 0]
                y = patch_grid[:, :, 1]

                # deal with boundaries
                x[x > max_x] = max_x
                y[y > max_y] = max_y
                x[x < 0] = 0
                y[y < 0] = 0

                kernel_response = response_image[x, y, j] * self._kernel_grid
                normalizer = np.sum(kernel_response)
                normalized_kernel_response = kernel_response / normalizer

                mean_shift_target[j, :] = np.sum(
                    normalized_kernel_response * (x, y), axis=(1, 2))

            # Compute (shape) error term
            error = mean_shift_target - target.points

            # Compute steepest descent parameter updates
            sd_delta_p = np.einsum('ijk, ik -> j', self._J, error)

            # TODO: a similar approach could be implemented in LK
            # Deal with prior
            prior = self._J_prior * self.transform.as_vector()

            # Compute parameter updates
            delta_p = -np.dot(self._inv_H, prior - sd_delta_p)

            # Update transform weights
            parameters = self.transform.as_vector() + delta_p
            fitting_result.parameters.append(parameters)
            self.transform.from_vector_inplace(parameters)
            target = self.transform.target

            # Test convergence
            error = np.abs(np.linalg.norm(delta_p))
            n_iters += 1

        fitting_result.fitted = True
        return fitting_result



from base import menpo_src_dir_path, Vectorizable, Targetable
from base import DP, DX, DL

import fit
import fitmultilevel
import image
import io
import landmark
import math
import model
import rasterize
import shape
import transform
import visualize

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np

from menpo.base import DX, DL

from .base import Transform, Alignment, Invertible
from .rbf import R2LogR2RBF


# Note we inherit from Alignment first to get it's n_dims behavior
class ThinPlateSplines(Alignment, Transform, Invertible, DX, DL):
    r"""
    The thin plate splines (TPS) alignment between 2D source and target
    landmarks.

    `kernel` can be used to specify an alternative kernel function. If
    `None` is supplied, the :class:`menpo.basis.rbf.R2LogR2` kernel will be
    used.

    Parameters
    ----------
    source : (N, 2) ndarray
        The source points to apply the tps from
    target : (N, 2) ndarray
        The target points to apply the tps to
    kernel : :class:`menpo.basis.rbf.BasisFunction`, optional
        The kernel to apply.

        Default: :class:`menpo.basis.rbf.R2LogR2`

    Raises
    ------
    ValueError
        TPS is only with on 2-dimensional data
    """

    def __init__(self, source, target, kernel=None):
        Alignment.__init__(self, source, target)
        if self.n_dims != 2:
            raise ValueError('TPS can only be used on 2D data.')
        if kernel is None:
            kernel = R2LogR2RBF(source.points)
        self.kernel = kernel
        # k[i, j] is the rbf weighting between source i and j
        # (of course, k is thus symmetrical and it's diagonal nil)
        self.k = self.kernel.apply(self.source.points)
        # p is a homogeneous version of the source points
        self.p = np.concatenate(
            [np.ones([self.n_points, 1]), self.source.points], axis=1)
        o = np.zeros([3, 3])
        top_l = np.concatenate([self.k, self.p], axis=1)
        bot_l = np.concatenate([self.p.T, o], axis=1)
        self.l = np.concatenate([top_l, bot_l], axis=0)
        self.v, self.y, self.coefficients = None, None, None
        self._build_coefficients()

    def _build_coefficients(self):
        self.v = self.target.points.T.copy()
        self.y = np.hstack([self.v, np.zeros([2, 3])])
        self.coefficients = np.linalg.solve(self.l, self.y.T)

    def _sync_state_from_target(self):
        # now the target is updated, we only have to rebuild the
        # coefficients.
        self._build_coefficients()

    def _apply(self, points, **kwargs):
        """
        Performs a TPS transform on the given points.

        Parameters
        ----------
        points : (N, D) ndarray
            The points to transform.

        Returns
        --------
        f : (N, D) ndarray
            The transformed points
        """
        if points.shape[1] != self.n_dims:
            raise ValueError('TPS can only be applied to 2D data.')
        x = points[..., 0][:, None]
        y = points[..., 1][:, None]
        # calculate the affine coefficients of the warp
        # (C = Constant component, then X, Y respectively)
        c_affine_c = self.coefficients[-3]
        c_affine_x = self.coefficients[-2]
        c_affine_y = self.coefficients[-1]
        # the affine warp component
        f_affine = c_affine_c + c_affine_x * x + c_affine_y * y
        # calculate a distance matrix (for L2 Norm) between every source
        # and the target
        kernel_dist = self.kernel.apply(points)
        # grab the affine free components of the warp
        c_affine_free = self.coefficients[:-3]
        # build the affine free warp component
        f_affine_free = kernel_dist.dot(c_affine_free)
        return f_affine + f_affine_free

    @property
    def has_true_inverse(self):
        return False

    def _build_pseudoinverse(self):
        return ThinPlateSplines(self.target, self.source, kernel=self.kernel)

    def d_dx(self, points):
        r"""
        The first order derivative of this TPS warp wrt spatial changes
        evaluated at points.

        Parameters
        ----------
        points: ndarray shape (n_points, n_dims)
            The spatial points at which the derivative should be evaluated.

        Returns
        -------
        d_dx: ndarray shape (n_points, n_dims, n_dims)
            The jacobian wrt spatial changes.

            d_dx[i, j, k] is the scalar differential change that the
            j'th dimension of the i'th point experiences due to a first order
            change in the k'th dimension.

        """
        dk_dx = np.zeros((points.shape[0] + 3,   # i
                          self.source.n_points,  # k
                          self.source.n_dims))   # l
        dk_dx[:-3, :] = self.kernel.d_dl(points)

        affine_derivative = np.array([[0, 0],
                                      [1, 0],
                                      [0, 1]])
        dk_dx[-3:, :] = affine_derivative[:, None]

        return np.einsum('ij, ikl -> klj', self.coefficients, dk_dx)

    def d_dl(self, points):
        """
        Calculates the Jacobian of the TPS warp wrt to the source landmarks
        assuming that he target is equal to the source. This is a special
        case of the Jacobian wrt to the source landmarks that is used in AAMs
        to weight the relative importance of each pixel in the reference
        frame wrt to each one of the source landmarks.

        dW_dl =      dOmega_dl         *  k(points)
              = T *     d_L**-1_dl     *  k(points)
              = T * -L**-1 dL_dl L**-1 *  k(points)

        # per point
        (c, d) = (d, c+3) (c+3, c+3) (c+3, c+3, c, d) (c+3, c+3) (c+3)
        (c, d) = (d,            c+3) (c+3, c+3, c, d) (c+3,)
        (c, d) = (d,               ) (          c, d)
        (c, d) = (                 ) (          c, d)

        Parameters
        ----------
        points : (n_points, n_dims)
            Points at which the Jacobian will be evaluated.

        Returns
        -------
        dW/dl : (n_points, n_params, n_dims) ndarray
            The Jacobian of the transform wrt to the source landmarks evaluated
            at the previous points and assuming that the target is equal to
            the source.
        """
        n_centres = self.n_points
        n_points = points.shape[0]

        # TPS kernel (nonlinear + affine)

        # for each input, evaluate the rbf
        # (n_points, n_centres)
        k_points = self.kernel.apply(points)

        # k_points with (1, x, y) appended to each point
        # (n_points, n_centres+3)  - 3 is (1, x, y) for affine component
        k = np.hstack([k_points, np.ones([n_points, 1]), points])

        # (n_centres+3, n_centres+3)
        inv_L = np.linalg.inv(self.l)

        # Taking the derivative of L for changes in l must yield an x,y change
        # for each centre.
        # (n_centres+3, n_centres+3, n_centres, n_dims)
        dL_dl = np.zeros(self.l.shape + (n_centres, 2))

        # take the derivative of the kernel wrt centres at the centres
        # SHOULD be (n_centres, n_dims, n_centres, n_dims)
        # IS        (n_centres,         n_centres, n_dims
        dK_dl_at_tgt = self.kernel.d_dl(self.source.points)

        # we want to build a tensor where for each slice where
        # dK_dl[i, j, k, l] is the derivative wrt the l'th dimension of the
        # i'th centre for L[j, k] -> first axis is just looping over centres
        # and last looping over dims
        # (n_centres, n_centres, n_centres, n_dims)
        dK_dl = np.zeros((n_centres, ) + dK_dl_at_tgt.shape)

        # make a linear iterator over the centres
        iter = np.arange(n_centres)

        # efficiently build the repeated pattern for dK_dl
        # note that the repetition over centres happens over axis 0
        # and the dims axis is the last
        # so dK_dl[0, ..., 0] corresponds to dK/dx0 in Joan's paper
        #    dK_dl[3, ..., 1] corresponds to dK_dy3 in Joan's paper
        dK_dl[iter, iter] = dK_dl_at_tgt[iter]
        dK_dl[iter, :, iter] = dK_dl_at_tgt[:, iter]

        # prepare memory for the answer
        # SHOULD be (n_points, n_dims, n_centres, n_dims)
        # IS        (n_points,       , n_centres, n_dims)
        dW_dl = np.zeros((n_points, n_centres, 2))

        # pretend the target is equal to the source
        # (n_dims, n_centres+3)
        pseudo_target = np.hstack([self.source.points.T, np.zeros([2, 3])])

        for i in np.arange(n_centres):
            # dP_dli (n_centres, n_points, n_dims, n_dims)
            dP_dli = np.zeros(self.p.shape + (2,))
            dP_dli[i, 1, 0] = -1
            dP_dli[i, 2, 1] = -1

            dL_dl[:n_centres, :n_centres, i] = dK_dl[i]
            dL_dl[:n_centres, n_centres:, i] = dP_dli
            dL_dl[n_centres:, :n_centres, i] = np.swapaxes(dP_dli, 0, 1)

            omega_x = -inv_L.dot(dL_dl[..., i, 0].dot(inv_L))
            omega_y = -inv_L.dot(dL_dl[..., i, 1].dot(inv_L))
            dW_dl[:, i, 0] = k.dot(omega_x).dot(pseudo_target[0])
            dW_dl[:, i, 1] = k.dot(omega_y).dot(pseudo_target[1])

        return dW_dl



import numpy as np

from menpo.image import Image
from menpo.fitmultilevel.base import MultilevelFitter
from menpo.fitmultilevel.featurefunctions import compute_features
from menpo.fitmultilevel.aam.base import AAMFitter
from menpo.fitmultilevel.clm.base import CLMFitter


class SDFitter(MultilevelFitter):
    r"""
    Mixin for Supervised Descent Fitters.
    """
    def _set_up(self):
        r"""
        Sets up the SD fitter object.
        """
        pass

    def fit(self, image, initial_shape, max_iters=None, gt_shape=None,
            error_type='me_norm', verbose=False, view=False, **kwargs):
        r"""
        Fits a single image.

        Parameters
        -----------
        image : :map:`MaskedImage`
            The image to be fitted.
        initial_shape : :map:`PointCloud`
            The initial shape estimate from which the fitting procedure
            will start.
        max_iters :  int  or `list`, optional
            The maximum number of iterations.

            If `int`, then this will be the overall maximum number of iterations
            for all the pyramidal levels.

            If `list`, then a maximum number of iterations is specified for each
            pyramidal level.

        gt_shape : :map:`PointCloud`
            The groundtruth shape of the image.

        error_type : {'me_norm', 'me', 'rmse'}, optional.
            Specifies the way in which the error between the fitted and
            ground truth shapes is to be computed.

        verbose: `boolean`, optional
            Whether or not to print information related to the fitting
            results (such as: ``final error``, ``convergence``, ...).

        view : `boolean`, optional
            Whether or not the fitting results are to be displayed.

        **kwargs : `dict`
            optional arguments to be passed through.

        Returns
        -------
        fitting_list : :map:`FittingResultList`
            A fitting result object.
        """
        if max_iters is None:
            max_iters = self.n_levels
        return MultilevelFitter.fit(self, image, initial_shape,
                                    max_iters=max_iters, gt_shape=gt_shape,
                                    error_type=error_type, verbose=verbose,
                                    view=view, **kwargs)


class SDMFitter(SDFitter):
    r"""
    Supervised Descent Method.

    Parameters
    -----------
    regressors : :map:`RegressorTrainer`
        The trained regressors.
        
    n_training_images : `int`
        The number of images that were used to train the SDM fitter. It is
        only used for informational reasons.
        
    feature_type : ``None`` or `string` or `function` or list of those, optional
        If list of length ``n_levels``, then a feature is defined per level.
        However, this requires that the ``pyramid_on_features`` flag is
        ``False``, so that the features are extracted at each level.
        The first element of the list specifies the features to be extracted at
        the lowest pyramidal level and so on.

        If not a list or a list with length ``1``, then:
            If ``pyramid_on_features`` is ``True``, the specified feature will
            be applied to the highest level.
            If ``pyramid_on_features`` is ``False``, the specified feature will
            be applied to all pyramid levels.

        Per level:
            If ``None``, the appearance model will be built using the original image
            representation, i.e. no features will be extracted from the original
            images.

            If `string`, image features will be computed by executing::

               feature_image = getattr(image.features, feature_type[level])()

            for each pyramidal level. For this to work properly each string
            needs to be one of menpo's standard image feature methods
            ('igo', 'hog', ...).
            Note that, in this case, the feature computation will be
            carried out using the default options.

        Non-default feature options and new experimental features can be
        defined using `function`. In this case, the `function` must
        receive an image as input and return a particular feature
        representation of that image. For example::

            def igo_double_from_std_normalized_intensities(image)
                image = deepcopy(image)
                image.normalize_std_inplace()
                return image.feature_type.igo(double_angles=True)

        See :map:`ImageFeatures` for details more details on
        Menpo's standard image features and feature options.

    reference_shape : :map:`PointCloud`
        The reference shape that was used to resize all training images to a
        consistent object size.
        
    downscale : `float`
        The downscale factor that will be used to create the different
        pyramidal levels. The scale factor will be::

            (downscale ** k) for k in range(n_levels)
            
    pyramid_on_features : `boolean`, optional
        If ``True``, the feature space is computed once at the highest scale and
        the Gaussian pyramid is applied on the feature images.

        If ``False``, the Gaussian pyramid is applied on the original images
        (intensities) and then features will be extracted at each level.
        
    interpolator : `string`
        The interpolator that was used during training.

    References
    ----------
    .. [XiongD13] Supervised Descent Method and its Applications to
       Face Alignment
       Xuehan Xiong and Fernando De la Torre Fernando
       IEEE International Conference on Computer Vision and Pattern Recognition
       May, 2013
    """
    def __init__(self, regressors, n_training_images, feature_type,
                 reference_shape, downscale, pyramid_on_features,
                 interpolator):
        self._fitters = regressors
        self._feature_type = feature_type
        self._reference_shape = reference_shape
        self._downscale = downscale
        self._interpolator = interpolator
        self._pyramid_on_features = pyramid_on_features
        self._n_training_images = n_training_images

    @property
    def algorithm(self):
        r"""
        Returns a string containing the algorithm used from the SDM family.

        : str
        """
        return 'SDM-' + self._fitters[0].algorithm

    @property
    def reference_shape(self):
        r"""
        The reference shape used during training.

        :type: :map:`PointCloud`
        """
        return self._reference_shape

    @property
    def feature_type(self):
        r"""
        The feature type per pyramid level. Note that they are stored from
        lowest to highest level resolution.

        :type: `list`
        """
        return self._feature_type

    @property
    def n_levels(self):
        r"""
        The number of pyramidal levels used during training.

        : int
        """
        return len(self._fitters)

    @property
    def downscale(self):
        r"""
        The downscale per pyramidal level used during building the AAM.
        The scale factor is: (downscale ** k) for k in range(n_levels)

        :type: `float`
        """
        return self._downscale

    @property
    def pyramid_on_features(self):
        r"""
        Flag that controls the Gaussian pyramid of the testing image based on
        the pyramid used during building.
        If True, the feature space is computed once at the highest scale and
        the Gaussian pyramid is applied on the feature images.
        If False, the Gaussian pyramid is applied on the original images
        (intensities) and then features will be extracted at each level.

        :type: `boolean`
        """
        return self._pyramid_on_features

    @property
    def interpolator(self):
        r"""
        The employed interpolator.

        :type: `string`
        """
        return self._interpolator

    def __str__(self):
        out = "Supervised Descent Method\n" \
              " - Non-Parametric '{}' Regressor\n" \
              " - {} training images.\n".format(
              self._fitters[0].regressor.__name__, self._n_training_images)
        # small strings about number of channels, channels string and downscale
        down_str = []
        for j in range(self.n_levels):
            if j == self.n_levels - 1:
                down_str.append('(no downscale)')
            else:
                down_str.append('(downscale by {})'.format(
                    self.downscale**(self.n_levels - j - 1)))
        temp_img = Image(image_data=np.random.rand(40, 40))
        if self.pyramid_on_features:
            temp = compute_features(temp_img, self.feature_type[0])
            n_channels = [temp.n_channels] * self.n_levels
        else:
            n_channels = []
            for j in range(self.n_levels):
                temp = compute_features(temp_img, self.feature_type[j])
                n_channels.append(temp.n_channels)
        # string about features and channels
        if self.pyramid_on_features:
            if isinstance(self.feature_type[0], str):
                feat_str = "- Feature is {} with ".format(
                    self.feature_type[0])
            elif self.feature_type[0] is None:
                feat_str = "- No features extracted. "
            else:
                feat_str = "- Feature is {} with ".format(
                    self.feature_type[0].func_name)
            if n_channels[0] == 1:
                ch_str = ["channel"]
            else:
                ch_str = ["channels"]
        else:
            feat_str = []
            ch_str = []
            for j in range(self.n_levels):
                if isinstance(self.feature_type[j], str):
                    feat_str.append("- Feature is {} with ".format(
                        self.feature_type[j]))
                elif self.feature_type[j] is None:
                    feat_str.append("- No features extracted. ")
                else:
                    feat_str.append("- Feature is {} with ".format(
                        self.feature_type[j].func_name))
                if n_channels[j] == 1:
                    ch_str.append("channel")
                else:
                    ch_str.append("channels")
        if self.n_levels > 1:
            out = "{} - Gaussian pyramid with {} levels and downscale " \
                  "factor of {}.\n".format(out, self.n_levels,
                                           self.downscale)
            if self.pyramid_on_features:
                out = "{}   - Pyramid was applied on feature space.\n   " \
                      "{}{} {} per image.\n".format(out, feat_str,
                                                    n_channels[0], ch_str[0])
            else:
                out = "{}   - Features were extracted at each pyramid " \
                      "level.\n".format(out)
                for i in range(self.n_levels - 1, -1, -1):
                    out = "{}   - Level {} {}: \n     {}{} {} per " \
                          "image.\n".format(
                          out, self.n_levels - i, down_str[i], feat_str[i],
                          n_channels[i], ch_str[i])
        else:
            if self.pyramid_on_features:
                feat_str = [feat_str]
            out = "{0} - No pyramid used:\n   {1}{2} {3} per image.\n".format(
                  out, feat_str[0], n_channels[0], ch_str[0])
        return out


class SDAAMFitter(AAMFitter, SDFitter):
    r"""
    Supervised Descent Fitter for AAMs.

    Parameters
    -----------
    aam : :map:`AAM`
        The Active Appearance Model to be used.

    regressors : :map:``RegressorTrainer`
        The trained regressors.

    n_training_images : `int`
        The number of training images used to train the SDM fitter.
    """
    def __init__(self, aam, regressors, n_training_images):
        super(SDAAMFitter, self).__init__(aam)
        self._fitters = regressors
        self._n_training_images = n_training_images

    @property
    def algorithm(self):
        r"""
        Returns a string containing the algorithm used from the SDM family.

        :type: `string`
        """
        return 'SD-AAM-' + self._fitters[0].algorithm

    def __str__(self):
        return "{}Supervised Descent Method for AAMs:\n" \
               " - Parametric '{}' Regressor\n" \
               " - {} training images.\n".format(
               self.aam.__str__(), self._fitters[0].regressor.__name__,
               self._n_training_images)


class SDCLMFitter(CLMFitter, SDFitter):
    r"""
    Supervised Descent Fitter for CLMs.

    Parameters
    -----------
    clm : :map:`CLM`
        The Constrained Local Model to be used.

    regressors : :map:`RegressorTrainer`
        The trained regressors.

    n_training_images : `int`
        The number of training images used to train the SDM fitter.

    References
    ----------
    .. [Asthana13] Robust Discriminative Response Map Fitting with Constrained
       Local Models
       A. Asthana, S. Zafeiriou, S. Cheng, M. Pantic.
       IEEE Conference onComputer Vision and Pattern Recognition.
       Portland, Oregon, USA, June 2013.
    """
    def __init__(self, clm, regressors, n_training_images):
        super(SDCLMFitter, self).__init__(clm)
        self._fitters = regressors
        self._n_training_images = n_training_images

    @property
    def algorithm(self):
        r"""
        Returns a string containing the algorithm used from the SDM family.

        :type: `string`
        """
        return 'SD-CLM-' + self._fitters[0].algorithm

    def __str__(self):
        return "{}Supervised Descent Method for CLMs:\n" \
               " - Parametric '{}' Regressor\n" \
               " - {} training images.\n".format(
               self.clm.__str__(), self._fitters[0].regressor.__name__,
               self._n_training_images)



from base import (import_auto, import_image, import_images,
                  import_mesh, import_meshes, import_builtin_asset,
                  import_landmark_file, import_landmark_files,
                  data_path_to, data_dir_path, ls_builtin_assets,
                  mesh_paths, image_paths)



import numpy as np

from .base import HomogFamilyAlignment
from .affine import Affine


class Similarity(Affine):
    r"""
    Specialist version of an :map:`Affine` that is guaranteed to be
    a Similarity transform.

    Parameters
    ----------
    h_matrix : (D + 1, D + 1) ndarray
        The homogeneous matrix of the similarity transform.

    """

    def __init__(self, h_matrix, copy=True, skip_checks=False):
        Affine.__init__(self, h_matrix, copy=copy, skip_checks=skip_checks)

    def _transform_str(self):
        r"""
        A string representation explaining what this similarity transform does.

        Returns
        -------
        str : string
            String representation of transform.

        """
        header = 'Similarity decomposing into:'
        list_str = [t._transform_str() for t in self.decompose()]
        return header + reduce(lambda x, y: x + '\n' + '  ' + y, list_str, '  ')

    @property
    def h_matrix_is_mutable(self):
        return False

    @classmethod
    def identity(cls, n_dims):
        return Similarity(np.eye(n_dims + 1))

    @property
    def n_parameters(self):
        r"""
        2D Similarity: 4 parameters::

            [(1 + a), -b,      tx]
            [b,       (1 + a), ty]

        3D Similarity: Currently not supported

        Returns
        -------
        int

        Raises
        ------
        DimensionalityError, NotImplementedError
            Only 2D transforms are supported.

        """
        if self.n_dims == 2:
            return 4
        elif self.n_dims == 3:
            raise NotImplementedError("3D similarity transforms cannot be "
                                      "vectorized yet.")
        else:
            raise ValueError("Only 2D and 3D Similarity transforms "
                             "are currently supported.")

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. The parameters
        are output in the order `[a, b, tx, ty]`, given that
        `a = k cos(theta) - 1` and `b = k sin(theta)` where `k` is a
        uniform scale and `theta` is a clockwise rotation in radians.

        **2D**

        ========= ===========================================
        parameter definition
        ========= ===========================================
        a         `a = k cos(theta) - 1`
        b         `b = k sin(theta)`
        tx        Translation in `x`
        ty        Translation in `y`
        ========= ===========================================

        .. note::

            Only 2D transforms are currently supported.

        Returns
        -------
        params : (P,) ndarray
            The values that parameterise the transform.

        Raises
        ------
        DimensionalityError, NotImplementedError
            If the transform is not 2D

        """
        n_dims = self.n_dims
        if n_dims == 2:
            params = self.h_matrix - np.eye(n_dims + 1)
            # Pick off a, b, tx, ty
            params = params[:n_dims, :].ravel(order='F')
            # Pick out a, b, tx, ty
            return params[[0, 1, 4, 5]]
        elif n_dims == 3:
            raise NotImplementedError("3D similarity transforms cannot be "
                                      "vectorized yet.")
        else:
            raise ValueError("Only 2D and 3D Similarity transforms "
                             "are currently supported.")

    def from_vector_inplace(self, p):
        r"""
        Returns an instance of the transform from the given parameters,
        expected to be in Fortran ordering.

        Supports rebuilding from 2D parameter sets.

        2D Similarity: 4 parameters::

            [a, b, tx, ty]

        Parameters
        ----------
        p : (P,) ndarray
            The array of parameters.

        Raises
        ------
        DimensionalityError, NotImplementedError
            Only 2D transforms are supported.

        """
        if p.shape[0] == 4:
            homog = np.eye(3)
            homog[0, 0] += p[0]
            homog[1, 1] += p[0]
            homog[0, 1] = -p[1]
            homog[1, 0] = p[1]
            homog[:2, 2] = p[2:]
            self._set_h_matrix(homog, skip_checks=True, copy=False)
        elif p.shape[0] == 7:
            raise NotImplementedError("3D similarity transforms cannot be "
                                      "vectorized yet.")
        else:
            raise ValueError("Only 2D and 3D Similarity transforms "
                             "are currently supported.")

    def _build_pseudoinverse(self):
        return Similarity(np.linalg.inv(self.h_matrix), copy=False,
                          skip_checks=True)

    def d_dp(self, points):
        r"""
        Computes the Jacobian of the transform w.r.t the parameters.

        The Jacobian generated (for 2D) is of the form::

            x -y 1 0
            y  x 0 1

        This maintains a parameter order of::

          W(x;p) = [1 + a  -b   ] [x] + tx
                   [b      1 + a] [y] + ty

        Parameters
        ----------
        points : (n_points, n_dims) ndarray
            The set of points to calculate the jacobian for.

        Returns
        -------
        (n_points, n_params, n_dims) ndarray
            The jacobian wrt parametrization

        Raises
        ------
        DimensionalityError
            `points.n_dims != self.n_dims` or transform is not 2D

        """
        n_points, points_n_dim = points.shape
        if points_n_dim != self.n_dims:
            raise ValueError('Trying to sample jacobian in incorrect '
                             'dimensions (transform is {0}D, sampling '
                             'at {1}D)'.format(self.n_dims, points_n_dim))
        elif self.n_dims != 2:
            # TODO: implement 3D Jacobian
            raise ValueError("Only the Jacobian of a 2D similarity "
                             "transform is currently supported.")

        # prealloc the jacobian
        jac = np.zeros((n_points, self.n_parameters, self.n_dims))
        ones = np.ones_like(points)

        # Build a mask and apply it to the points to build the jacobian
        # Do this for each parameter - [a, b, tx, ty] respectively
        self._apply_jacobian_mask(jac, np.array([1, 1]), 0, points)
        self._apply_jacobian_mask(jac, np.array([-1, 1]), 1, points[:, ::-1])
        self._apply_jacobian_mask(jac, np.array([1, 0]), 2, ones)
        self._apply_jacobian_mask(jac, np.array([0, 1]), 3, ones)

        return jac

    def _apply_jacobian_mask(self, jac, param_mask, row_index, points):
        # make a mask for a single points jacobian
        full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=np.bool)
        # fill the mask in for the ith axis
        full_mask[row_index] = [True, True]
        # assign the ith axis points to this mask, broadcasting over all
        # points
        jac[:, full_mask] = points * param_mask


class AlignmentSimilarity(HomogFamilyAlignment, Similarity):
    """
    Infers the similarity transform relating two vectors with the same
    dimensionality. This is simply the procrustes alignment of the
    source to the target.

    Parameters
    ----------

    source : :map:`PointCloud`
        The source pointcloud instance used in the alignment

    target : :map:`PointCloud`
        The target pointcloud instance used in the alignment

    rotation: boolean, optional
        If False, the rotation component of the similarity transform is not
        inferred.

        Default: True

    """
    def __init__(self, source, target, rotation=True):
        HomogFamilyAlignment.__init__(self, source, target)
        x = self._procrustes_alignment(source, target, rotation=rotation)
        Similarity.__init__(self, x.h_matrix, copy=False, skip_checks=True)

    @staticmethod
    def _procrustes_alignment(source, target, rotation=True):
        r"""
        Returns the similarity transform that aligns the source to the target.

        """
        from .rotation import Rotation, optimal_rotation_matrix
        from .translation import Translation
        from .scale import UniformScale
        target_translation = Translation(-target.centre)
        centred_target = target_translation.apply(target)
        # now translate the source to the origin
        translation = Translation(-source.centre)
        # apply the translation to the source
        aligned_source = translation.apply(source)
        scale = UniformScale(target.norm() / source.norm(), source.n_dims)
        scaled_aligned_source = scale.apply(aligned_source)
        # compute the target's inverse translation
        inv_target_translation = target_translation.pseudoinverse
        if rotation:
            rotation = Rotation(optimal_rotation_matrix(scaled_aligned_source,
                                                        centred_target))
            return translation.compose_before(scale).compose_before(
                rotation).compose_before(inv_target_translation)
        else:
            return translation.compose_before(scale).compose_before(
                inv_target_translation)

    def _sync_state_from_target(self):
        similarity = self._procrustes_alignment(self.source, self.target)
        self._set_h_matrix(similarity.h_matrix, copy=False, skip_checks=True)

    def as_non_alignment(self):
        r"""Returns a copy of this similarity without it's alignment nature.

        Returns
        -------
        transform : :map:`Similarity`
            A version of this similarity with the same transform behavior but
            without the alignment logic.
        """
        return Similarity(self.h_matrix, skip_checks=True)

    def from_vector_inplace(self, p):
        r"""
        Returns an instance of the transform from the given parameters,
        expected to be in Fortran ordering.

        Supports rebuilding from 2D parameter sets.

        2D Similarity: 4 parameters::

            [a, b, tx, ty]

        Parameters
        ----------
        p : (P,) ndarray
            The array of parameters.

        Raises
        ------
        DimensionalityError, NotImplementedError
            Only 2D transforms are supported.

        """
        Similarity.from_vector_inplace(self, p)
        self._sync_target_from_state()



import abc


class Fitter(object):
    r"""
    Interface that all Fitter objects must implement.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def algorithm(self):
        r"""
        Returns the name of the fitter object.
        """
        pass

    @abc.abstractmethod
    def _set_up(self, **kwargs):
        r"""
        Abstract method that sets up the fitter object.
        """
        pass

    def fit(self, image, initial_parameters, gt_shape=None, **kwargs):
        r"""
        Fits the fitter object to an image.

        Parameters
        -----------
        image: :class:`menpo.image.masked.MaskedImage`
            The image to be fitted.
        initial_parameters: list
            The initial parameters of the model.
        gt_shape: :class:`menpo.shape.PointCloud`, optional
            The original ground truth shape associated to the image.

            Default: None

        Returns
        -------
        fitting_result: `menpo.fit.fittingresult`
            The fitting result.
        """
        fitting_result = self._create_fitting_result(
            image, initial_parameters, gt_shape=gt_shape)
        return self._fit(fitting_result, **kwargs)

    @abc.abstractmethod
    def _create_fitting_result(self, **kwargs):
        r"""
        Abstract method that defines the fitting result object associated to
        the fitter object.
        """
        pass

    @abc.abstractmethod
    def _fit(self, **kwargs):
        r"""
        Abstract method implements a particular alignment algorithm.
        """
        pass

    def get_parameters(self, shape):
        r"""
        Abstract method that gets the parameters.
        """
        pass



from .predefined import (aam_fastest_alternating_noise,
                         aam_fastest_alternating_bbox,
                         aam_best_performance_alternating_noise,
                         aam_best_performance_alternating_bbox,
                         aam_params_combinations_noise,
                         clm_params_combinations_noise, clm_basic_noise,
                         clm_basic_bbox)
from .io import download_ibug_bounding_boxes



from __future__ import division
import numpy as np

from menpo.shape import PointCloud
from menpo.transform import Similarity, AlignmentSimilarity


# TODO: document me
def build_sampling_grid(patch_shape):
    r"""
    """
    patch_shape= np.array(patch_shape)
    patch_half_shape = np.require(np.round(patch_shape / 2), dtype=int)
    start = -patch_half_shape
    end = patch_half_shape + 1
    sampling_grid = np.mgrid[start[0]:end[0], start[1]:end[1]]
    return sampling_grid.swapaxes(0, 2).swapaxes(0, 1)


# TODO: document me
def extract_local_patches(image, centres, sampling_grid):
    r"""
    Extract square patches from an image about centres.

    Parameters
    ----------
    image : :map:`Image`
        The image to extract patches from

    centres : :map:`PointCloud`
        The centres around which the patches should be extracted

    sampling_grid : `ndarray` of shape: ``(patch_shape[0]+1, patch_shape[1]+1,
    n_dims)``
        The size of the patch in each dimension

    Returns
    -------
    patches : ndarray` of shape: ``n_centres, + patch_shape + n_channels,``
        The patches as a single numpy array.
    """
    max_x = image.shape[0] - 1
    max_y = image.shape[1] - 1

    pixels = image.pixels

    patch_grid = (sampling_grid[None, ...] +
                  np.round(centres.points[:, None, None, ...]).astype(int))

    X = patch_grid[:, :, :, 0]
    Y = patch_grid[:, :, :, 1]

    # deal with boundaries
    X[X > max_x] = max_x
    Y[Y > max_y] = max_y
    X[X < 0] = 0
    Y[Y < 0] = 0

    patches = np.empty(
        (centres.n_points,) + sampling_grid.shape[:-1] + (image.n_channels,))
    for j, (x, y) in enumerate(zip(X, Y)):
        patches[j, ...] = pixels[x, y, :]

    return patches


def extract_local_patches_fast(image, centres, patch_shape, out=None):
    r"""
    Extract square patches from an image about centres.

    Parameters
    ----------
    image : :map:`Image`
        The image to extract patches from

    centres : :map:`PointCloud`
        The centres around which the patches should be extracted

    patch_shape : `tuple` of `ints`
        The size of the patch in each dimension

    out : `ndarray` of shape: ``n_centres, + patch_shape + n_channels,``,
    optional
        The output array to be assigned to. If `None`, a new numpy array
        will be created.

    Returns
    -------
    patches : `ndarray` of shape: ``n_centres, + patch_shape + n_channels,``
        The patches as a single numpy array.
    """
    if out is not None:
        patches = out
    else:
        patches = np.empty(
            (centres.n_points,) + patch_shape + (image.n_channels,))
    # 0 out the patches array
    patches[...] = 0
    image_size = np.array(image.shape, dtype=np.int)
    patch_shape = np.array(patch_shape, dtype=np.int)
    centres = np.require(centres.points, dtype=np.int)
    half_patch_shape = np.require(np.ceil(patch_shape / 2), dtype=np.int)
    # 1. compute the extents
    c_min = centres - half_patch_shape
    c_max = centres + half_patch_shape
    out_min_min = c_min < 0
    out_min_max = c_min > image_size
    out_max_min = c_max < 0
    out_max_max = c_max > image_size

    # 1. Build the extraction slices
    ext_s_min = c_min.copy()
    ext_s_max = c_max.copy()
    # Clamp the min to 0
    ext_s_min[out_min_min] = 0
    ext_s_max[out_max_min] = 0
    # Clamp the max to image bounds across each dimension
    for i in xrange(image.n_dims):
        ext_s_max[out_max_max[:, i], i] = image_size[i] - 1
        ext_s_min[out_min_max[:, i], i] = image_size[i] - 1

    # 2. Build the insertion slices
    ins_s_min = ext_s_min - c_min
    ins_s_max = np.maximum(ext_s_max - c_max + patch_shape, (0, 0))

    for i, (e_a, e_b, i_a, i_b) in enumerate(zip(ext_s_min, ext_s_max,
                                                 ins_s_min, ins_s_max)):
        # build a list of insertion slices and extraction slices
        i_slices = [slice(a, b) for a, b in zip(i_a, i_b)]
        e_slices = [slice(a, b) for a, b in zip(e_a, e_b)]
        # get a view onto the patch we are on
        patch = patches[i, ...]
        # apply the slices to map
        patch[i_slices] = image.pixels[e_slices]

    return patches


def mean_pointcloud(pointclouds):
    r"""
    Compute the mean of a list of point cloud objects

    Parameters
    ----------
    pointclouds: list of :class:`menpo.shape.PointCloud`
        List of point cloud objects from which we want to
        compute the mean.

    Returns
    -------
    mean_pointcloud: class:`menpo.shape.PointCloud`
        The mean point cloud.
    """
    return PointCloud(np.mean([pc.points for pc in pointclouds], axis=0))


# TODO: Should this be a method on Similarity? AlignableTransforms?
def noisy_align(source, target, noise_std=0.04, rotation=False):
    r"""
    Constructs and perturbs the optimal similarity transform between source
    to the target by adding white noise to its weights.

    Parameters
    ----------
    source: :class:`menpo.shape.PointCloud`
        The source pointcloud instance used in the alignment
    target: :class:`menpo.shape.PointCloud`
        The target pointcloud instance used in the alignment
    noise_std: float
        The standard deviation of the white noise

        Default: 0.04
    rotation: boolean
        If False the second parameter of the Similarity,
        which captures captures inplane rotations, is set to 0.

        Default:False

    Returns
    -------
    noisy_transform : :class: `menpo.transform.Similarity`
        The noisy Similarity Transform
    """
    transform = AlignmentSimilarity(source, target, rotation=rotation)
    parameters = transform.as_vector()
    parameter_range = np.hstack((parameters[:2], target.range()))
    noise = (parameter_range * noise_std *
             np.random.randn(transform.n_parameters))
    return Similarity.identity(source.n_dims).from_vector(parameters + noise)


def align_shape_with_bb(shape, bounding_box):
    r"""
    Returns the Similarity transform that aligns the provided shape with the
    provided bounding box.

    Parameters
    ----------
    shape: :class:`menpo.shape.PointCloud`
        The shape to be aligned.
    bounding_box: (2, 2) ndarray
        The bounding box specified as:

            np.array([[x_min, y_min], [x_max, y_max]])

    Returns
    -------
    transform : :class: `menpo.transform.Similarity`
        The align transform
    """
    shape_box = PointCloud(shape.bounds())
    bounding_box = PointCloud(bounding_box)
    return AlignmentSimilarity(shape_box, bounding_box, rotation=False)


#TODO: Document me
def compute_error(target, ground_truth, error_type='me_norm'):
    r"""
    """
    if error_type == 'me_norm':
        return _compute_me_norm(target, ground_truth)
    elif error_type == 'me':
        return _compute_me(target, ground_truth)
    elif error_type == 'rmse':
        return _compute_rmse(target, ground_truth)
    else:
        raise ValueError("Unknown error_type string selected. Valid options "
                         "are: me_norm, me, rmse'")


#TODO: Document me
def _compute_rmse(target, ground_truth):
    r"""
    """
    return np.sqrt(np.mean((target.flatten() - ground_truth.flatten()) ** 2))


#TODO: Document me
def _compute_me(target, ground_truth):
    r"""
    """
    return np.mean(np.sqrt(np.sum((target - ground_truth) ** 2, axis=-1)))


#TODO: Document me
def _compute_me_norm(target, ground_truth):
    r"""
    """
    normalizer = np.mean(np.max(ground_truth, axis=0) -
                         np.min(ground_truth, axis=0))
    return _compute_me(target, ground_truth) / normalizer



from __future__ import division, print_function
import numpy as np

from menpo.image import Image
from menpo.fitmultilevel.builder import DeformableModelBuilder
from menpo.fitmultilevel.functions import build_sampling_grid
from menpo.fitmultilevel.featurefunctions import compute_features, sparse_hog
from menpo.visualize import print_dynamic, progress_bar_str

from .classifierfunctions import classifier, linear_svm_lr


class CLMBuilder(DeformableModelBuilder):
    r"""
    Class that builds Multilevel Constrained Local Models.

    Parameters
    ----------
    classifier_type : ``classifier_closure`` or list of those
        If list of length ``n_levels``, then a classifier function is defined
        per level. The first element of the list specifies the classifier to be
        used at the lowest pyramidal level and so on.

        If not a list or a list with length ``1``, then the specified classifier
        function will be used for all levels.

        Per level:
             A closure implementing a binary classifier.

        Examples of such closures can be found in
        :ref:`clm_builders`

    patch_shape : tuple of `int`
        The shape of the patches used by the previous classifier closure.

    feature_type : ``None`` or `string` or `function` or list of those, optional
        If list of length ``n_levels``, then a feature is defined per level.
        However, this requires that the ``pyramid_on_features`` flag is
        disabled, so that the features are extracted at each level.
        The first element of the list specifies the features to be extracted
        at the lowest pyramidal level and so on.

        If not a list or a list with length ``1``, then:
            If ``pyramid_on_features`` is ``True``, the specified feature will
            be applied to the highest level.
            If ``pyramid_on_features`` is ``False``, the specified feature will
            be applied to all pyramid levels.

        Per level:
            If ``None``, the appearance model will be built using the original
            image representation, i.e. no features will be extracted from the
            original images.

        If `string`, image features will be computed by executing::

           feature_image = getattr(image.features, feature_type[level])()

        for each pyramidal level. For this to work properly each `string`
        needs to be one of Menpo's standard image feature methods
        ('igo', 'hog', ...).
        Note that, in this case, the feature computation will be
        carried out using the default options.

        Non-default feature options and new experimental features can be
        defined using functions. In this case, the functions must
        receive an image as input and return a particular feature
        representation of that image. For example::

            def igo_double_from_std_normalized_intensities(image)
                image = deepcopy(image)
                image.normalize_std_inplace()
                return image.feature_type.igo(double_angles=True)

        See :map:`ImageFeatures` for details more details on
        Menpo's standard image features and feature options.

    normalization_diagonal : `int` >= ``20``, optional
        During building an AAM, all images are rescaled to ensure that the
        scale of their landmarks matches the scale of the mean shape.

        If `int`, it ensures that the mean shape is scaled so that the diagonal
        of the bounding box containing it matches the ``normalization_diagonal``
        value.
        If ``None``, the mean shape is not rescaled.

        Note that, because the reference frame is computed from the mean
        landmarks, this kwarg also specifies the diagonal length of the
        reference frame (provided that features computation does not change
        the image size).

    n_levels : `int` > ``0``, optional
        The number of multi-resolution pyramidal levels to be used.

    downscale : `float` >= ``1``, optional
        The downscale factor that will be used to create the different
        pyramidal levels. The scale factor will be::

            (downscale ** k) for k in range(n_levels)

    scaled_shape_models : `boolean`, optional
        If ``True``, the reference frames will be the mean shapes of each
        pyramid level, so the shape models will be scaled.

        If ``False``, the reference frames of all levels will be the mean shape
        of the highest level, so the shape models will not be scaled; they will
        have the same size.

    pyramid_on_features : `boolean`, optional
        If ``True``, the feature space is computed once at the highest scale and
        the Gaussian pyramid is applied on the feature images.

        If ``False``, the Gaussian pyramid is applied on the original images
        (intensities) and then features will be extracted at each level.

    max_shape_components : ``None`` or `int` > ``0`` or ``0`` <= `float` <= ``1`` or list of those, optional
        If list of length ``n_levels``, then a number of shape components is
        defined per level. The first element of the list specifies the number
        of components of the lowest pyramidal level and so on.

        If not a list or a list with length ``1``, then the specified number of
        shape components will be used for all levels.

        Per level:
            If `int`, it specifies the exact number of components to be
            retained.

            If `float`, it specifies the percentage of variance to be retained.

            If ``None``, all the available components are kept
            (100% of variance).

    boundary : `int` >= ``0``, optional
        The number of pixels to be left as a safe margin on the boundaries
        of the reference frame (has potential effects on the gradient
        computation).

    interpolator : `string`, optional
        The interpolator that should be used to perform the warps.

    Returns
    -------
    clm : :map:`CLMBuilder`
        The CLM Builder object
    """
    def __init__(self, classifier_type=linear_svm_lr, patch_shape=(5, 5),
                 feature_type=sparse_hog, normalization_diagonal=None,
                 n_levels=3, downscale=1.1, scaled_shape_models=True,
                 pyramid_on_features=False, max_shape_components=None,
                 boundary=3, interpolator='scipy'):
        # check parameters
        self.check_n_levels(n_levels)
        self.check_downscale(downscale)
        self.check_normalization_diagonal(normalization_diagonal)
        self.check_boundary(boundary)
        max_shape_components = self.check_max_components(
            max_shape_components, n_levels, 'max_shape_components')
        feature_type = self.check_feature_type(feature_type, n_levels,
                                               pyramid_on_features)
        classifier_type = check_classifier_type(classifier_type, n_levels)
        patch_shape = check_patch_shape(patch_shape)

        # store parameters
        self.classifier_type = classifier_type
        self.patch_shape = patch_shape
        self.feature_type = feature_type
        self.normalization_diagonal = normalization_diagonal
        self.n_levels = n_levels
        self.downscale = downscale
        self.scaled_shape_models = scaled_shape_models
        self.pyramid_on_features = pyramid_on_features
        self.max_shape_components = max_shape_components
        self.boundary = boundary
        self.interpolator = interpolator

    def build(self, images, group=None, label='all', verbose=False):
        r"""
        Builds a Multilevel Constrained Local Model from a list of
        landmarked images.

        Parameters
        ----------
        images : list of :map:`Image`
            The set of landmarked images from which to build the AAM.

        group : string, Optional
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`, optional
            The label of of the landmark manager that you wish to use. If
            ``None``, the convex hull of all landmarks is used.

        verbose : `boolean`, optional
            Flag that controls information and progress printing.

        Returns
        -------
        clm : :map:`CLM`
            The CLM object
        """
        # compute reference_shape and normalize images size
        self.reference_shape, normalized_images = \
            self._normalization_wrt_reference_shape(
                images, group, label, self.normalization_diagonal,
                self.interpolator, verbose=verbose)

        # create pyramid
        generators = self._create_pyramid(normalized_images, self.n_levels,
                                          self.downscale,
                                          self.pyramid_on_features,
                                          self.feature_type, verbose=verbose)

        # build the model at each pyramid level
        if verbose:
            if self.n_levels > 1:
                print_dynamic('- Building model for each of the {} pyramid '
                              'levels\n'.format(self.n_levels))
            else:
                print_dynamic('- Building model\n')

        shape_models = []
        classifiers = []
        # for each pyramid level (high --> low)
        for j in range(self.n_levels):
            # since models are built from highest to lowest level, the
            # parameters of type list need to use a reversed index
            rj = self.n_levels - j - 1

            if verbose:
                level_str = '  - '
                if self.n_levels > 1:
                    level_str = '  - Level {}: '.format(j + 1)

            # get images of current level
            feature_images = []
            if self.pyramid_on_features:
                # features are already computed, so just call generator
                for c, g in enumerate(generators):
                    if verbose:
                        print_dynamic('{}Rescaling feature space - {}'.format(
                            level_str,
                            progress_bar_str((c + 1.) / len(generators),
                                             show_bar=False)))
                    feature_images.append(g.next())
            else:
                # extract features of images returned from generator
                for c, g in enumerate(generators):
                    if verbose:
                        print_dynamic('{}Computing feature space - {}'.format(
                            level_str,
                            progress_bar_str((c + 1.) / len(generators),
                                             show_bar=False)))
                    feature_images.append(compute_features(
                        g.next(), self.feature_type[rj]))

            # extract potentially rescaled shapes
            shapes = [i.landmarks[group][label].lms for i in feature_images]

            # define shapes that will be used for training
            if j == 0:
                original_shapes = shapes
                train_shapes = shapes
            else:
                if self.scaled_shape_models:
                    train_shapes = shapes
                else:
                    train_shapes = original_shapes

            # train shape model and find reference frame
            if verbose:
                print_dynamic('{}Building shape model'.format(level_str))
            shape_model = self._build_shape_model(
                train_shapes, self.max_shape_components[rj])

            # add shape model to the list
            shape_models.append(shape_model)

            # build classifiers
            sampling_grid = build_sampling_grid(self.patch_shape)
            n_points = shapes[0].n_points
            level_classifiers = []
            for k in range(n_points):
                if verbose:
                    print_dynamic('{}Building classifiers - {}'.format(
                        level_str,
                        progress_bar_str((k + 1.) / n_points,
                                         show_bar=False)))

                positive_labels = []
                negative_labels = []
                positive_samples = []
                negative_samples = []

                for i, s in zip(feature_images, shapes):

                    max_x = i.shape[0] - 1
                    max_y = i.shape[1] - 1

                    point = (np.round(s.points[k, :])).astype(int)
                    patch_grid = sampling_grid + point[None, None, ...]
                    positive, negative = get_pos_neg_grid_positions(
                        patch_grid, positive_grid_size=(1, 1))

                    x = positive[:, 0]
                    y = positive[:, 1]
                    x[x > max_x] = max_x
                    y[y > max_y] = max_y
                    x[x < 0] = 0
                    y[y < 0] = 0

                    positive_sample = i.pixels[positive[:, 0],
                                               positive[:, 1], :]
                    positive_samples.append(positive_sample)
                    positive_labels.append(np.ones(positive_sample.shape[0]))

                    x = negative[:, 0]
                    y = negative[:, 1]
                    x[x > max_x] = max_x
                    y[y > max_y] = max_y
                    x[x < 0] = 0
                    y[y < 0] = 0

                    negative_sample = i.pixels[x, y, :]
                    negative_samples.append(negative_sample)
                    negative_labels.append(-np.ones(negative_sample.shape[0]))

                positive_samples = np.asanyarray(positive_samples)
                positive_samples = np.reshape(positive_samples,
                                              (-1, positive_samples.shape[-1]))
                positive_labels = np.asanyarray(positive_labels).flatten()

                negative_samples = np.asanyarray(negative_samples)
                negative_samples = np.reshape(negative_samples,
                                              (-1, negative_samples.shape[-1]))
                negative_labels = np.asanyarray(negative_labels).flatten()

                X = np.vstack((positive_samples, negative_samples))
                t = np.hstack((positive_labels, negative_labels))

                clf = classifier(X, t, self.classifier_type[rj])
                level_classifiers.append(clf)

            # add level classifiers to the list
            classifiers.append(level_classifiers)

            if verbose:
                print_dynamic('{}Done\n'.format(level_str))

        # reverse the list of shape and appearance models so that they are
        # ordered from lower to higher resolution
        shape_models.reverse()
        classifiers.reverse()
        n_training_images = len(images)

        return CLM(shape_models, classifiers, n_training_images,
                   self.patch_shape, self.feature_type, self.reference_shape,
                   self.downscale, self.scaled_shape_models,
                   self.pyramid_on_features, self.interpolator)


class CLM(object):
    r"""
    Constrained Local Model class.

    Parameters
    -----------
    shape_models : :map:`PCAModel` list
        A list containing the shape models of the CLM.

    classifiers : ``classifier_closure`` list of lists
        A list containing the list of classifier_closures per each pyramidal
        level of the CLM.

    n_training_images : `int`
        The number of training images used to build the AAM.

    patch_shape : tuple of `int`
        The shape of the patches used to train the classifiers.

    feature_type : ``None`` or `string` or `function` or list of those
        The image feature that was be used to build the ``appearance_models``.
        Will subsequently be used by fitter objects using this class to fit to
        novel images.

        If list of length ``n_levels``, then a feature was defined per level.
        This means that the ``pyramid_on_features`` flag was ``False``
        and the features were extracted at each level. The first element of
        the list specifies the features of the lowest pyramidal level and so
        on.

        If not a list or a list with length ``1``, then:
            If ``pyramid_on_features`` is ``True``, the specified feature
            was applied to the highest level.

            If ``pyramid_on_features`` is ``False``, the specified feature was
            applied to all pyramid levels.

        Per level:
            If ``None``, the appearance model was built using the original
            image representation, i.e. no features will be extracted from the
            original images.

            If `string`, the appearance model was built using one of Menpo's
            default built-in feature representations - those
            accessible at ``image.features.some_feature()``. Note that this case
            can only be used with default feature parameters - for custom
            feature weights, use the functional form of this argument instead.

            If `function`, the user can directly provide the feature that was
            calculated on the images. This class will simply invoke this
            function, passing in as the sole argument the image to be fitted,
            and expect as a return type an :map:`Image` representing the feature
            calculation ready for further fitting. See the examples for
            details.

    reference_shape : :map:`PointCloud`
        The reference shape that was used to resize all training images to a
        consistent object size.

    downscale : `float`
        The downscale factor that was used to create the different pyramidal
        levels.

    scaled_shape_models : `boolean`, Optional
        If ``True``, the reference frames are the mean shapes of each pyramid
        level, so the shape models are scaled.

        If ``False``, the reference frames of all levels are the mean shape of
        the highest level, so the shape models are not scaled; they have the
        same size.

    pyramid_on_features : `boolean`, optional
        If True, the feature space was computed once at the highest scale and
        the Gaussian pyramid was applied on the feature images.
        If False, the Gaussian pyramid was applied on the original images
        (intensities) and then features were extracted at each level.

    interpolator : `string`
        The interpolator that was used to build the CLM.
    """
    def __init__(self, shape_models, classifiers, n_training_images,
                 patch_shape, feature_type, reference_shape, downscale,
                 scaled_shape_models, pyramid_on_features, interpolator):
        self.shape_models = shape_models
        self.classifiers = classifiers
        self.n_training_images = n_training_images
        self.patch_shape = patch_shape
        self.feature_type = feature_type
        self.reference_shape = reference_shape
        self.downscale = downscale
        self.scaled_shape_models = scaled_shape_models
        self.pyramid_on_features = pyramid_on_features
        self.interpolator = interpolator

    @property
    def n_levels(self):
        """
        The number of multi-resolution pyramidal levels of the CLM.

        :type: `int`
        """
        return len(self.shape_models)

    @property
    def n_classifiers_per_level(self):
        """
        The number of classifiers per pyramidal level of the CLM.

        :type: `int`
        """
        return [len(clf) for clf in self.classifiers]

    def instance(self, shape_weights=None, level=-1):
        r"""
        Generates a novel CLM instance given a set of shape weights. If no
        weights are provided, the mean CLM instance is returned.

        Parameters
        -----------
        shape_weights : ``(n_weights,)`` `ndarray` or `float` list
            Weights of the shape model that will be used to create
            a novel shape instance. If `None`, the mean shape
            ``(shape_weights = [0, 0, ..., 0])`` is used.

        level : `int`, optional
            The pyramidal level to be used.

        Returns
        -------
        shape_instance : :map:`PointCloud`
            The novel CLM instance.
        """
        sm = self.shape_models[level]
        # TODO: this bit of logic should to be transferred down to PCAModel
        if shape_weights is None:
            shape_weights = [0]
        n_shape_weights = len(shape_weights)
        shape_weights *= sm.eigenvalues[:n_shape_weights] ** 0.5
        shape_instance = sm.instance(shape_weights)
        return shape_instance

    def random_instance(self, level=-1):
        r"""
        Generates a novel random CLM instance.

        Parameters
        -----------
        level : `int`, optional
            The pyramidal level to be used.

        Returns
        -------
        shape_instance : :map:`PointCloud`
            The novel CLM instance.
        """
        sm = self.shape_models[level]
        # TODO: this bit of logic should to be transferred down to PCAModel
        shape_weights = (np.random.randn(sm.n_active_components) *
                         sm.eigenvalues[:sm.n_active_components]**0.5)
        shape_instance = sm.instance(shape_weights)
        return shape_instance

    def response_image(self, image, group=None, label='all', level=-1):
        r"""
        Generates a response image result of applying the classifiers of a
        particular pyramidal level of the CLM to an image.

        Parameters
        -----------
        image: :map:`Image`
            The image.

        group : `string`, optional
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        label : `string`, optional
            The label of of the landmark manager that you wish to use. If no
            label is passed, the convex hull of all landmarks is used.

        level: `int`, optional
            The pyramidal level to be used.

        Returns
        -------
        image : :map:`Image`
            The response image.
        """
        # rescale image
        image = image.rescale_to_reference_shape(self.reference_shape,
                                                 group=group, label=label)

        # apply pyramid
        if self.n_levels > 1:
            if self.pyramid_on_features:
                # compute features at highest level
                feature_image = compute_features(image, self.feature_type[0])

                # apply pyramid on feature image
                pyramid = feature_image.gaussian_pyramid(
                    n_levels=self.n_levels, downscale=self.downscale)

                # get rescaled feature images
                images = list(pyramid)
            else:
                # create pyramid on intensities image
                pyramid = image.gaussian_pyramid(
                    n_levels=self.n_levels, downscale=self.downscale)

                # compute features at each level
                images = [compute_features(
                    i, self.feature_type[self.n_levels - j - 1])
                    for j, i in enumerate(pyramid)]
            images.reverse()
        else:
            images = [compute_features(image, self.feature_type[0])]

        # initialize responses
        image = images[level]
        image_pixels = np.reshape(image.pixels, (-1, image.n_channels))
        response_data = np.zeros((image.shape[0], image.shape[1],
                                  self.n_classifiers_per_level[level]))
        # Compute responses
        for j, clf in enumerate(self.classifiers[level]):
            response_data[:, :, j] = np.reshape(clf(image_pixels),
                                                image.shape)
        return Image(image_data=response_data)

    @property
    def _str_title(self):
        r"""
        Returns a string containing name of the model.

        : str
        """
        return 'Constrained Local Model'

    def __str__(self):
        out = "{}\n - {} training images.\n".format(self._str_title,
                                                    self.n_training_images)
        # small strings about number of channels, channels string and downscale
        down_str = []
        for j in range(self.n_levels):
            if j == self.n_levels - 1:
                down_str.append('(no downscale)')
            else:
                down_str.append('(downscale by {})'.format(
                    self.downscale**(self.n_levels - j - 1)))
        temp_img = Image(image_data=np.random.rand(50, 50))
        if self.pyramid_on_features:
            temp = compute_features(temp_img, self.feature_type[0])
            n_channels = [temp.n_channels] * self.n_levels
        else:
            n_channels = []
            for j in range(self.n_levels):
                temp = compute_features(temp_img, self.feature_type[j])
                n_channels.append(temp.n_channels)
        # string about features and channels
        if self.pyramid_on_features:
            if isinstance(self.feature_type[0], str):
                feat_str = "- Feature is {} with ".format(
                    self.feature_type[0])
            elif self.feature_type[0] is None:
                feat_str = "- No features extracted. "
            else:
                feat_str = "- Feature is {} with ".format(
                    self.feature_type[0].func_name)
            if n_channels[0] == 1:
                ch_str = ["channel"]
            else:
                ch_str = ["channels"]
        else:
            feat_str = []
            ch_str = []
            for j in range(self.n_levels):
                if isinstance(self.feature_type[j], str):
                    feat_str.append("- Feature is {} with ".format(
                        self.feature_type[j]))
                elif self.feature_type[j] is None:
                    feat_str.append("- No features extracted. ")
                else:
                    feat_str.append("- Feature is {} with ".format(
                        self.feature_type[j].func_name))
                if n_channels[j] == 1:
                    ch_str.append("channel")
                else:
                    ch_str.append("channels")
        if self.n_levels > 1:
            if self.scaled_shape_models:
                out = "{} - Gaussian pyramid with {} levels and downscale " \
                      "factor of {}.\n   - Each level has a scaled shape " \
                      "model (reference frame).\n   - Patch size is {}W x " \
                      "{}H.\n".format(out, self.n_levels, self.downscale,
                                      self.patch_shape[1], self.patch_shape[0])

            else:
                out = "{} - Gaussian pyramid with {} levels and downscale " \
                      "factor of {}:\n   - Shape models (reference frames) " \
                      "are not scaled.\n   - Patch size is {}W x " \
                      "{}H.\n".format(out, self.n_levels, self.downscale,
                                      self.patch_shape[1], self.patch_shape[0])
            if self.pyramid_on_features:
                out = "{}   - Pyramid was applied on feature space.\n   " \
                      "{}{} {} per image.\n".format(out, feat_str,
                                                    n_channels[0], ch_str[0])
            else:
                out = "{}   - Features were extracted at each pyramid " \
                      "level.\n".format(out)
            for i in range(self.n_levels - 1, -1, -1):
                out = "{}   - Level {} {}: \n".format(out, self.n_levels - i,
                                                      down_str[i])
                if not self.pyramid_on_features:
                    out = "{}     {}{} {} per image.\n".format(
                        out, feat_str[i], n_channels[i], ch_str[i])
                out = "{0}     - {1} shape components ({2:.2f}% of " \
                      "variance)\n     - {3} {4} classifiers.\n".format(
                      out, self.shape_models[i].n_components,
                      self.shape_models[i].variance_ratio * 100,
                      self.n_classifiers_per_level[i],
                      self.classifiers[i][0].func_name)
        else:
            if self.pyramid_on_features:
                feat_str = [feat_str]
            out = "{0} - No pyramid used:\n   {1}{2} {3} per image.\n" \
                  "   - {4} shape components ({5:.2f}% of " \
                  "variance)\n   - {6} {7} classifiers.".format(
                  out, feat_str[0], n_channels[0], ch_str[0],
                  self.shape_models[0].n_components,
                  self.shape_models[0].variance_ratio * 100,
                  self.n_classifiers_per_level[0],
                  self.classifiers[0][0].func_name)
        return out


def get_pos_neg_grid_positions(sampling_grid, positive_grid_size=(1, 1)):
    r"""
    Divides a sampling grid in positive and negative pixel positions. By
    default only the center of the grid is considered to be positive.
    """
    positive_grid_size = np.array(positive_grid_size)
    mask = np.zeros(sampling_grid.shape[:-1], dtype=np.bool)
    center = np.round(np.array(mask.shape) / 2).astype(int)
    positive_grid_size -= [1, 1]
    start = center - positive_grid_size
    end = center + positive_grid_size + 1
    mask[start[0]:end[0], start[1]:end[1]] = True
    positive = sampling_grid[mask]
    negative = sampling_grid[~mask]
    return positive, negative


def check_classifier_type(classifier_type, n_levels):
    r"""
    Checks the classifier type per level. It must be a classifier
    function closure or a list containing 1 or {n_levels} closures.
    """
    str_error = ("classifier_type must be a classifier function closure "
                 "of a list containing 1 or {} closures").format(n_levels)
    if not isinstance(classifier_type, list):
        classifier_type_list = [classifier_type] * n_levels
    elif len(classifier_type) == 1:
        classifier_type_list = [classifier_type[0]] * n_levels
    elif len(classifier_type) == n_levels:
        classifier_type_list = classifier_type
    else:
        raise ValueError(str_error)
    for clas in classifier_type_list:
        if not hasattr(clas, '__call__'):
            raise ValueError(str_error)
    return classifier_type_list


def check_patch_shape(patch_shape):
    r"""
    Checks the patch shape. It must be a tuple with `int` > ``1``.
    """
    str_error = "patch_size mast be a tuple with two integers"
    if not isinstance(patch_shape, tuple) or len(patch_shape) != 2:
        raise ValueError(str_error)
    for sh in patch_shape:
        if not isinstance(sh, int) or sh < 2:
            raise ValueError(str_error)
    return patch_shape



import os
import numpy as np

import menpo.io as mio
from menpo.visualize.text_utils import print_dynamic, progress_bar_str
from menpo.fitmultilevel.aam import AAMBuilder, LucasKanadeAAMFitter
from menpo.fitmultilevel.clm import CLMBuilder, GradientDescentCLMFitter
from menpo.fitmultilevel.sdm import SDMTrainer, SDMFitter
from menpo.fit.fittingresult import FittingResultList
from menpo.landmark import labeller
from menpo.visualize.base import GraphPlotter


def aam_fit_benchmark(fitting_images, aam, fitting_options=None,
                      perturb_options=None, verbose=False):
    r"""
    Fits a trained AAM model to a database.

    Parameters
    ----------
    fitting_images: list of :class:MaskedImage objects
        A list of the fitting images.
    aam: :class:menpo.fitmultilevel.aam.AAM object
        The trained AAM object. It can be generated from the
        aam_build_benchmark() method.
    fitting_options: dictionary, optional
        A dictionary with the parameters that will be passed in the
        LucasKanadeAAMFitter (:class:menpo.fitmultilevel.aam.base).
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            fitting_options = {'algorithm': AlternatingInverseCompositional,
                               'md_transform': OrthoMDTransform,
                               'global_transform': AlignmentSimilarity,
                               'n_shape': None,
                               'n_appearance': None,
                               'max_iters': 50,
                               'error_type': 'me_norm'
                               }
        For an explanation of the options, please refer to the
        LucasKanadeAAMFitter documentation.

        Default: None
    bounding_boxes: list of (2, 2) ndarray, optional
        If provided, fits will be initialized from a bounding box. If
        None, perturbation of ground truth will be used instead.
        can be provided). Interpreted as [[min_x, min_y], [max_x, max_y]].
    perturb_options: dictionary, optional
        A dictionary with parameters that control the perturbation on the
        ground truth shape with noise of specified std. Note that if
        bounding_box is provided perturb_options is ignored and not used.
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            initialization_options = {'noise_std': 0.04,
                                      'rotation': False
                                      }
        For an explanation of the options, please refer to the perturb_shape()
        method documentation of :map:`MultilevelFitter`.
    verbose: bool, optional
        If True, it prints information regarding the AAM fitting including
        progress bar, current image error and percentage of images with errors
        less or equal than a value.

        Default: False

    Returns
    -------
    fitting_results: :map:`FittingResultList`
        A list with the :map:`FittingResult` object per image.
    """
    if verbose:
        print('AAM Fitting:')
        perc1 = 0.
        perc2 = 0.

    # parse options
    if fitting_options is None:
        fitting_options = {}
    if perturb_options is None:
        perturb_options = {}

    # extract some options
    group = fitting_options.pop('gt_group', 'PTS')
    max_iters = fitting_options.pop('max_iters', 50)
    error_type = fitting_options.pop('error_type', 'me_norm')

    # create fitter
    fitter = LucasKanadeAAMFitter(aam, **fitting_options)

    # fit images
    n_images = len(fitting_images)
    fitting_results = []
    for j, i in enumerate(fitting_images):
        # perturb shape
        gt_s = i.landmarks[group].lms
        if 'bbox' in i.landmarks:
            # shape from bounding box
            s = fitter.obtain_shape_from_bb(i.landmarks['bbox'].lms.points)
        else:
            # shape from perturbation
            s = fitter.perturb_shape(gt_s, **perturb_options)
        # fit
        fr = fitter.fit(i, s, gt_shape=gt_s, max_iters=max_iters,
                        error_type=error_type, verbose=False)
        fitting_results.append(fr)

        # print
        if verbose:
            if error_type == 'me_norm':
                if fr.final_error <= 0.03:
                    perc1 += 1.
                if fr.final_error <= 0.04:
                    perc2 += 1.
            elif error_type == 'rmse':
                if fr.final_error <= 0.05:
                    perc1 += 1.
                if fr.final_error <= 0.06:
                    perc2 += 1.
            print_dynamic('- {0} - [<=0.03: {1:.1f}%, <=0.04: {2:.1f}%] - '
                          'Image {3}/{4} (error: {5:.3f} --> {6:.3f})'.format(
                progress_bar_str(float(j + 1) / n_images, show_bar=False),
                perc1 * 100. / n_images, perc2 * 100. / n_images, j + 1,
                n_images, fr.initial_error, fr.final_error))
    if verbose:
        print_dynamic('- Fitting completed: [<=0.03: {0:.1f}%, <=0.04: '
                      '{1:.1f}%]\n'.format(perc1 * 100. / n_images,
                                           perc2 * 100. / n_images))

    # fit images
    fitting_results_list = FittingResultList(fitting_results)

    return fitting_results_list


def aam_build_benchmark(training_images, training_options=None, verbose=False):
    r"""
    Builds an AAM model.

    Parameters
    ----------
    training_images: list of :class:MaskedImage objects
        A list of the training images.
    training_options: dictionary, optional
        A dictionary with the parameters that will be passed in the AAMBuilder
        (:class:menpo.fitmultilevel.aam.AAMBuilder).
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            training_options = {'group': 'PTS',
                                'feature_type': 'igo',
                                'transform': PiecewiseAffine,
                                'trilist': None,
                                'normalization_diagonal': None,
                                'n_levels': 3,
                                'downscale': 2,
                                'scaled_shape_models': True,
                                'pyramid_on_features': True,
                                'max_shape_components': None,
                                'max_appearance_components': None,
                                'boundary': 3,
                                'interpolator': 'scipy'
                                }
        For an explanation of the options, please refer to the AAMBuilder
        documentation.

        Default: None
    verbose: boolean, optional
        If True, it prints information regarding the AAM training.

        Default: False

    Returns
    -------
    aam: :class:menpo.fitmultilevel.aam.AAM object
        The trained AAM model.
    """
    if verbose:
        print('AAM Training:')

    # parse options
    if training_options is None:
        training_options = {}

    # group option
    group = training_options.pop('group', None)

    # trilist option
    trilist = training_options.pop('trilist', None)
    if trilist is not None:
        labeller(training_images[0], 'PTS', trilist)
        training_options['trilist'] = \
            training_images[0].landmarks[trilist.__name__].lms.trilist

    # build aam
    aam = AAMBuilder(**training_options).build(training_images, group=group,
                                               verbose=verbose)

    return aam


def clm_fit_benchmark(fitting_images, clm, fitting_options=None,
                      perturb_options=None, verbose=False):
    r"""
    Fits a trained CLM model to a database.

    Parameters
    ----------
    fitting_images: list of :class:MaskedImage objects
        A list of the fitting images.
    clm: :class:menpo.fitmultilevel.clm.CLM object
        The trained CLM object. It can be generated from the
        clm_build_benchmark() method.
    fitting_options: dictionary, optional
        A dictionary with the parameters that will be passed in the
        GradientDescentCLMFitter (:class:menpo.fitmultilevel.clm.base).
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            fitting_options = {'algorithm': RegularizedLandmarkMeanShift,
                               'pdm_transform': OrthoPDM,
                               'global_transform': AlignmentSimilarity,
                               'n_shape': None,
                               'max_iters': 50,
                               'error_type': 'me_norm'
                               }
        For an explanation of the options, please refer to the
        GradientDescentCLMFitter documentation.

        Default: None
    bounding_boxes: list of (2, 2) ndarray, optional
        If provided, fits will be initialized from a bounding box. If
        None, perturbation of ground truth will be used instead.
        can be provided). Interpreted as [[min_x, min_y], [max_x, max_y]].
    perturb_options: dictionary, optional
        A dictionary with parameters that control the perturbation on the
        ground truth shape with noise of specified std. Note that if
        bounding_box is provided perturb_options is ignored and not used.
    verbose: boolean, optional
        If True, it prints information regarding the AAM fitting including
        progress bar, current image error and percentage of images with errors
        less or equal than a value.

        Default: False

    Returns
    -------
    fitting_results: :class:menpo.fit.fittingresult.FittingResultList object
        A list with the FittingResult object per image.
    """
    if verbose:
        print('CLM Fitting:')
        perc1 = 0.
        perc2 = 0.

    # parse options
    if fitting_options is None:
        fitting_options = {}

    # extract some options
    group = fitting_options.pop('gt_group', 'PTS')
    max_iters = fitting_options.pop('max_iters', 50)
    error_type = fitting_options.pop('error_type', 'me_norm')

    # create fitter
    fitter = GradientDescentCLMFitter(clm, **fitting_options)

    # fit images
    n_images = len(fitting_images)
    fitting_results = []
    for j, i in enumerate(fitting_images):
        # perturb shape
        gt_s = i.landmarks[group].lms
        if 'bbox' in i.landmarks:
            # shape from bounding box
            s = fitter.obtain_shape_from_bb(i.landmarks['bbox'].lms.points)
        else:
            # shape from perturbation
            s = fitter.perturb_shape(gt_s, **perturb_options)
        # fit
        fr = fitter.fit(i, s, gt_shape=gt_s, max_iters=max_iters,
                        error_type=error_type, verbose=False)
        fitting_results.append(fr)

        # print
        if verbose:
            if error_type == 'me_norm':
                if fr.final_error <= 0.03:
                    perc1 += 1.
                if fr.final_error <= 0.04:
                    perc2 += 1.
            elif error_type == 'rmse':
                if fr.final_error <= 0.05:
                    perc1 += 1.
                if fr.final_error <= 0.06:
                    perc2 += 1.
            print_dynamic('- {0} - [<=0.03: {1:.1f}%, <=0.04: {2:.1f}%] - '
                          'Image {3}/{4} (error: {5:.3f} --> {6:.3f})'.format(
                          progress_bar_str(float(j + 1) / n_images,
                                           show_bar=False),
                          perc1 * 100. / n_images, perc2 * 100. / n_images,
                          j + 1, n_images, fr.initial_error, fr.final_error))
    if verbose:
        print_dynamic('- Fitting completed: [<=0.03: {0:.1f}%, <=0.04: '
                      '{1:.1f}%]\n'.format(perc1 * 100. / n_images,
                                           perc2 * 100. / n_images))

    # fit images
    fitting_results_list = FittingResultList(fitting_results)

    return fitting_results_list


def clm_build_benchmark(training_images, training_options=None, verbose=False):
    r"""
    Builds an CLM model.

    Parameters
    ----------
    training_images: list of :class:MaskedImage objects
        A list of the training images.
    training_options: dictionary, optional
        A dictionary with the parameters that will be passed in the CLMBuilder
        (:class:menpo.fitmultilevel.clm.CLMBuilder).
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            training_options = {'group': 'PTS',
                                'classifier_type': linear_svm_lr,
                                'patch_shape': (5, 5),
                                'feature_type': sparse_hog,
                                'normalization_diagonal': None,
                                'n_levels': 3,
                                'downscale': 1.1,
                                'scaled_shape_models': True,
                                'pyramid_on_features': True,
                                'max_shape_components': None,
                                'boundary': 3,
                                'interpolator': 'scipy'
                                }
        For an explanation of the options, please refer to the CLMBuilder
        documentation.

        Default: None
    verbose: boolean, optional
        If True, it prints information regarding the CLM training.

        Default: False

    Returns
    -------
    clm: :class:menpo.fitmultilevel.clm.CLM object
        The trained CLM model.
    """
    if verbose:
        print('CLM Training:')

    # parse options
    if training_options is None:
        training_options = {}

    # group option
    group = training_options.pop('group', None)

    # build aam
    aam = CLMBuilder(**training_options).build(training_images, group=group,
                                               verbose=verbose)

    return aam


def sdm_fit_benchmark(fitting_images, fitter, perturb_options=None,
                      fitting_options=None, verbose=False):
    r"""
    Fits a trained SDM to a database.

    Parameters
    ----------
    fitting_images: list of :class:MaskedImage objects
        A list of the fitting images.
    fitter: :map:`SDMFitter`
        The trained AAM object. It can be generated from the
        aam_build_benchmark() method.
    fitting_options: dictionary, optional
        A dictionary with the parameters that will be passed in the
        LucasKanadeAAMFitter (:class:menpo.fitmultilevel.sdm.base).
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            fitting_options = {'algorithm': AlternatingInverseCompositional,
                               'md_transform': OrthoMDTransform,
                               'global_transform': AlignmentSimilarity,
                               'n_shape': None,
                               'n_appearance': None,
                               'max_iters': 50,
                               'error_type': 'me_norm'
                               }
        For an explanation of the options, please refer to the
        LucasKanadeAAMFitter documentation.

        Default: None
    bounding_boxes: list of (2, 2) ndarray, optional
        If provided, fits will be initialized from a bounding box. If
        None, perturbation of ground truth will be used instead.
        can be provided). Interpreted as [[min_x, min_y], [max_x, max_y]].
    perturb_options: dictionary, optional
        A dictionary with parameters that control the perturbation on the
        ground truth shape with noise of specified std. Note that if
        bounding_box is provided perturb_options is ignored and not used.
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            initialization_options = {'noise_std': 0.04,
                                      'rotation': False
                                      }
        For an explanation of the options, please refer to the perturb_shape()
        method documentation of :map:`MultilevelFitter`.
    verbose: bool, optional
        If True, it prints information regarding the AAM fitting including
        progress bar, current image error and percentage of images with errors
        less or equal than a value.

        Default: False

    Returns
    -------
    fitting_results: :map:`FittingResultList`
        A list with the :map:`FittingResult` object per image.
    """
    if verbose:
        print('SDM Fitting:')
        perc1 = 0.
        perc2 = 0.

    # parse options
    if fitting_options is None:
        fitting_options = {}
    if perturb_options is None:
        perturb_options = {}

    # extract some options
    group = fitting_options.pop('gt_group', 'PTS')
    error_type = fitting_options.pop('error_type', 'me_norm')

    # fit images
    n_images = len(fitting_images)
    fitting_results = []
    for j, i in enumerate(fitting_images):
        # perturb shape
        gt_s = i.landmarks[group].lms
        if 'bbox' in i.landmarks:
            # shape from bounding box
            s = fitter.obtain_shape_from_bb(i.landmarks['bbox'].lms.points)
        else:
            # shape from perturbation
            s = fitter.perturb_shape(gt_s, **perturb_options)
        # fit
        fr = fitter.fit(i, s, gt_shape=gt_s, error_type=error_type,
                        verbose=False)
        fitting_results.append(fr)

        # print
        if verbose:
            if error_type == 'me_norm':
                if fr.final_error <= 0.03:
                    perc1 += 1.
                if fr.final_error <= 0.04:
                    perc2 += 1.
            elif error_type == 'rmse':
                if fr.final_error <= 0.05:
                    perc1 += 1.
                if fr.final_error <= 0.06:
                    perc2 += 1.
            print_dynamic('- {0} - [<=0.03: {1:.1f}%, <=0.04: {2:.1f}%] - '
                          'Image {3}/{4} (error: {5:.3f} --> {6:.3f})'.format(
                progress_bar_str(float(j + 1) / n_images, show_bar=False),
                perc1 * 100. / n_images, perc2 * 100. / n_images, j + 1,
                n_images, fr.initial_error, fr.final_error))
    if verbose:
        print_dynamic('- Fitting completed: [<=0.03: {0:.1f}%, <=0.04: '
                      '{1:.1f}%]\n'.format(perc1 * 100. / n_images,
                                           perc2 * 100. / n_images))

    # fit images
    fitting_results_list = FittingResultList(fitting_results)

    return fitting_results_list


def sdm_build_benchmark(training_images, training_options=None, verbose=False):
    r"""
    Builds an SDM model.

    Parameters
    ----------
    training_images: list of :class:MaskedImage objects
        A list of the training images.
    training_options: dictionary, optional
        A dictionary with the parameters that will be passed in the AAMBuilder
        (:class:menpo.fitmultilevel.aam.AAMBuilder).
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            training_options = {'group': 'PTS',
                                'feature_type': 'igo',
                                'transform': PiecewiseAffine,
                                'trilist': None,
                                'normalization_diagonal': None,
                                'n_levels': 3,
                                'downscale': 2,
                                'scaled_shape_models': True,
                                'pyramid_on_features': True,
                                'max_shape_components': None,
                                'max_appearance_components': None,
                                'boundary': 3,
                                'interpolator': 'scipy'
                                }
        For an explanation of the options, please refer to the AAMBuilder
        documentation.

        Default: None
    verbose: boolean, optional
        If True, it prints information regarding the AAM training.

        Default: False

    Returns
    -------
    aam: :class:menpo.fitmultilevel.aam.AAM object
        The trained AAM model.
    """
    if verbose:
        print('SDM Training:')

    # parse options
    if training_options is None:
        training_options = {}

    # group option
    group = training_options.pop('group', None)

    # build sdm
    sdm = SDMTrainer(**training_options).train(training_images, group=group,
                                               verbose=verbose)
    return sdm


def load_database(database_path, bounding_boxes=None,
                  db_loading_options=None, verbose=False):
    r"""
    Loads the database images, crops them and converts them.

    Parameters
    ----------
    database_path: str
        The path of the database images.
    db_loading_options: dictionary, optional
        A dictionary with options related to image loading.
        If None, the default options will be used.
        This is an example of the dictionary with the default options:
            training_options = {'crop_proportion': 0.1,
                                'convert_to_grey': True,
                                }

        crop_proportion (float) defines the additional padding to be added all
        around the landmarks bounds when the images are cropped. It is defined
        as a proportion of the landmarks' range.

        convert_to_grey (boolean)defines whether the images will be converted
        to greyscale.

        Default: None
    verbose: boolean, optional
        If True, it prints a progress percentage bar.

        Default: False

    Returns
    -------
    images: list of :class:MaskedImage objects
        A list of the loaded images.

    Raises
    ------
    ValueError
        Invalid path given
    ValueError
        No {files_extension} files in given path
    """
    # check input options
    if db_loading_options is None:
        db_loading_options = {}

    # check given path
    database_path = os.path.abspath(os.path.expanduser(database_path))
    if os.path.isdir(database_path) is not True:
        raise ValueError('Invalid path given')

    # create final path
    final_path = os.path.join(database_path, '*')

    # get options
    crop_proportion = db_loading_options.pop('crop_proportion', 0.5)
    convert_to_grey = db_loading_options.pop('convert_to_grey', True)

    # load images
    images = []
    for i in mio.import_images(final_path, verbose=verbose):
        # If we have bounding boxes then we need to make sure we crop to them!
        # If we don't crop to the bounding box then we might crop out part of
        # the image the bounding box belongs to.
        landmark_group_label = None
        if bounding_boxes is not None:
            fname = i.ioinfo.filename + i.ioinfo.extension
            landmark_group_label = 'bbox'
            i.landmarks[landmark_group_label] = bounding_boxes[fname].detector

        # crop image
        i.crop_to_landmarks_proportion_inplace(crop_proportion,
                                               group=landmark_group_label)

        # convert it to greyscale if needed
        if convert_to_grey and i.n_channels == 3:
            i = i.as_greyscale(mode='luminosity')

        # append it to the list
        images.append(i)
    if verbose:
        print("\nAssets loaded.")
    return images


def convert_fitting_results_to_ced(fitting_results, max_error_bin=0.05,
                                   bins_error_step=0.005):
    r"""
    Method that given a fitting_result object, it converts it to the
    cumulative error distribution values that can be used for plotting.

    Parameters
    ----------
    fitting_results: :class:menpo.fit.fittingresult.FittingResultList object
        A list with the FittingResult object per image.
    max_error_bin: float, Optional
        The maximum error of the distribution.

        Default: 0.05
    bins_error_step: float, Optional
        The sampling step of the distribution values.

        Default: 0.005

    Returns
    -------
    final_error_dist: list
        Cumulative distribution values of the final errors.
    initial_error_dist: list
        Cumulative distribution values of the initial errors.
    """
    error_bins = np.arange(0., max_error_bin + bins_error_step,
                           bins_error_step)
    final_error_dist = np.array(
        [float(np.sum(fitting_results.final_error <= k)) /
         len(fitting_results.final_error) for k in error_bins])
    initial_error_dist = np.array(
        [float(np.sum(fitting_results.initial_error <= k)) /
         len(fitting_results.final_error) for k in error_bins])
    return final_error_dist, initial_error_dist, error_bins


def plot_fitting_curves(x_axis, ceds, title, figure_id=None, new_figure=False,
                        x_label='Point-to-Point Normalized RMS Error',
                        y_limit=1, x_limit=0.05, legend=None, **kwargs):
    r"""
    Method that plots Cumulative Error Distributions in a single figure.

    Parameters
    ----------
    x_axis: ndarray
        The horizontal axis values (errors).
    ceds: list of ndarrays
        The vertical axis values (percentages).
    title: string
        The plot title.
    figure_id, Optional
        A figure handle.

        Default: None
    new_figure: boolean, Optional
        If True, a new figure window will be created.

        Default: False
    y_limit: float, Optional
        The maximum value of the vertical axis.

        Default: 1
    x_limit: float, Optional
        The maximum value of the vertical axis.

        Default: 0.05
    x_label: string
        The label of the horizontal axis.

        Default: 'Point-to-Point Normalized RMS Error'
    legend: list of strings or None
        The legend of the plot. If None, the legend will include an incremental
        number per curve.

        Default: None

    Returns
    -------
    final_error_dist: list
        Cumulative distribution values of the final errors.
    initial_error_dist: list
        Cumulative distribution values of the initial errors.
    """
    if legend is None:
        legend = [str(i + 1) for i in range(len(ceds))]
    y_label = 'Proportion of images'
    axis_limits = [0, x_limit, 0, y_limit]
    return GraphPlotter(figure_id, new_figure, x_axis, ceds, title=title,
                        legend=legend, x_label=x_label, y_label=y_label,
                        axis_limits=axis_limits).render(**kwargs)
